from typing import Optional

from .expander import ExpanderHardware
from .module import ModuleHardware
from .robot_brain import RobotBrain


class CanHardware(ModuleHardware):

    def __init__(self, robot_brain: RobotBrain, *,
                 expander: Optional[ExpanderHardware] = None,
                 name: str = 'can',
                 rx_pin: int = 32,
                 tx_pin: int = 33,
                 baud: int = 1_000_000) -> None:
        self.name = name
        lizard_code = f'{self.name} = {expander.name + "." if expander else ""}Can({rx_pin}, {tx_pin}, {baud})'
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code)



import datetime
import logging
from pathlib import Path

import rosys

from . import registry


class BackupSchedule:
    """The BackupSchedule module is responsible for backing up the persistence files every day at the specified time."""

    def __init__(self,
                 path: Path = Path('~/.rosys_backup'),
                 time: datetime.time = datetime.time(3, 0),
                 backup_count: int = 100) -> None:
        self.path = path.expanduser()
        self.path.mkdir(parents=True, exist_ok=True)
        self.time = time
        self.backup_count = backup_count
        self.log = logging.getLogger('rosys.persistence')

        rosys.on_repeat(self.backup, 60)

    def backup(self) -> None:
        """Backup the persistence files every day at the specified time."""
        now = datetime.datetime.now()
        filepath = self.path / f'{now:%Y-%m-%d}.json'
        if now.time() < self.time or filepath.exists():
            return
        self.log.info('Backing up persistence files to %s', filepath)
        registry.write_export(filepath)
        self.prune()

    def prune(self) -> None:
        """Delete old backups if there are more than the specified number."""
        while len(backups := sorted(self.path.glob('*.json'))) > self.backup_count:
            backups[0].unlink()



import logging
import os
from collections import deque
from typing import Optional

import serial
from nicegui import ui
from nicegui.events import ValueChangeEventArguments
from serial.serialutil import SerialException

from ... import rosys
from .communication import Communication


class SerialCommunication(Communication):
    """This module implements a communication via a serial device with a given baud rate.

    It contains a list of search paths for finding the serial device.
    """

    search_paths: list[str] = [
        '/dev/tty.SLAB_USBtoUART',
        '/dev/ttyTHS1',
        '/dev/ttyUSB0',
    ]

    def __init__(self, baud_rate: int = 115200) -> None:
        super().__init__()
        self.device_path = self.get_device_path()
        if self.device_path is None:
            raise FileNotFoundError('No serial port found')
        self.log.debug('connecting serial on %s with baud rate %s', self.device_path, baud_rate)
        self.serial = serial.Serial(self.device_path, baud_rate)
        self.buffer = ''
        self.undo_queue: deque[str] = deque(maxlen=100)
        self.redo_queue: deque[str] = deque(maxlen=100)

    @staticmethod
    def is_possible() -> bool:
        device_path = SerialCommunication.get_device_path()
        if device_path is None:
            return False
        try:
            with serial.Serial(device_path):
                return True  # Successfully opened the port
        except SerialException:
            return False

    @staticmethod
    def get_device_path() -> Optional[str]:
        for device_path in SerialCommunication.search_paths:
            if os.path.exists(device_path) and os.stat(device_path).st_gid > 0:
                return device_path
        return None

    def connect(self) -> None:
        if not self.serial.isOpen():
            self.serial.open()
            self.log.debug('reconnected serial on %s', self.device_path)

    def disconnect(self) -> None:
        if self.serial.isOpen():
            self.serial.close()
            self.log.debug('disconnected serial on %s', self.device_path)

    async def read(self) -> Optional[str]:
        if not self.serial.isOpen():
            return None
        self.buffer += self.serial.read_all().decode(errors='replace')
        if '\n' in self.buffer:
            line, self.buffer = self.buffer.split('\r\n', 1)
            self.log.debug('read: %s', line)
            return line
        return None

    async def send(self, msg: str) -> None:
        if not self.serial.isOpen():
            return
        self.serial.write(f'{msg}\n'.encode())
        self.log.debug('send: %s', msg)

    def debug_ui(self) -> None:
        super().debug_ui()

        async def input_enter() -> None:
            await self.send(command.value)
            while self.redo_queue:
                self.undo_queue.append(self.redo_queue.pop())
            self.undo_queue.append(command.value)
            command.value = ''

        def input_up() -> None:
            if self.undo_queue:
                if command.value:
                    self.redo_queue.append(command.value)
                command.value = self.undo_queue.pop()

        def input_down() -> None:
            if command.value:
                self.undo_queue.append(command.value)
            command.value = self.redo_queue.pop() if self.redo_queue else ''

        def toggle(e: ValueChangeEventArguments) -> None:
            if e.value:
                self.connect()
                rosys.notify('connected to Lizard')
            else:
                self.disconnect()
                rosys.notify('disconnected from Lizard')

        ui.switch('Serial Communication', value=self.serial.isOpen(), on_change=toggle)
        ui.switch('Serial Logging', value=self.log.getEffectiveLevel() <= logging.DEBUG,
                  on_change=lambda e: self.log.setLevel(logging.DEBUG if e.value else logging.INFO))
        command = ui.input('Serial Command')
        command.on('keydown.enter', input_enter)
        command.on('keydown.up', input_up)
        command.on('keydown.down', input_down)



from enum import Enum

from typing_extensions import Self


class ImageRotation(str, Enum):
    NONE: int = 0
    RIGHT: int = 90
    UPSIDE_DOWN: int = 180
    LEFT: int = 270

    @classmethod
    def from_degrees(cls, degrees: int) -> Self:
        for rotation in cls:
            if int(rotation.value) == degrees % 360:
                return rotation
        raise ValueError(f'invalid rotation angle: {degrees}')



class BmsMessage:

    def __init__(self, bytes_: list[int]) -> None:
        self.bytes = bytes_
        self.cursor = -1

    @property
    def address(self) -> int:
        return self.bytes[1]

    @property
    def status(self) -> int:
        return self.bytes[2]

    @property
    def length(self) -> int:
        return self.bytes[3]

    @property
    def content(self) -> list[int]:
        return self.bytes[4:-3]

    def __str__(self) -> str:
        address = {0x03: 'STATUS', 0x04: 'VOLTAGE', 0x05: 'VERSION'}[self.address]
        return f'{address}: {" ".join(hex(b) for b in self.content)}'

    def check(self):
        assert self.bytes[0] == 0xdd
        assert (self.bytes[-3] << 8) + self.bytes[-2] == 0xffff - sum(self.content) - self.length - self.status + 1
        assert self.bytes[-1] == 0x77

    def get1(self):
        self.cursor += 1
        return self.content[self.cursor]

    def get2(self):
        self.cursor += 2
        return (self.content[self.cursor - 1] << 8) + self.content[self.cursor]

    def get2_signed(self):
        value = self.get2()
        return value if value < 32768 else value - 65536

    def interpret(self) -> dict:
        d = {}
        if self.address == 0x03:
            d['total voltage'] = self.get2() / 100
            d['current'] = self.get2_signed() / 100
            d['residual capacity'] = self.get2() / 100
            d['nominal capacity'] = self.get2() / 100
            d['cycle life'] = self.get2()
            date = self.get2()
            d['product year'] = 2000 + (date >> 9)
            d['product month'] = (date >> 5) & 0x0F
            d['product day'] = date & 0x1F
            d['balance status'] = self.get2()
            d['balance status high'] = self.get2()
            protection = self.get2()
            protections = [
                'cell block over-voltage',
                'cell block under-voltage',
                'battery over-voltage',
                'battery under-voltage',
                'charging over-temp',
                'charging low-temp',
                'discharging over-temp',
                'discharging low-temp',
                'charging over-current',
                'discharging over-current',
                'short circuit',
                'fore-end IC error',
                'MOS software lock-in',
            ]
            d['protections'] = {p: protection & (1 << i) for i, p in enumerate(protections)}
            version = self.get1()
            d['BMS major version'] = version >> 4
            d['BMS minor version'] = version & 0xf
            d['capacity percent'] = self.get1()
            d['fet'] = ["off", "charging", "discharging", "both"][self.get1()]
            d['number of blocks'] = self.get1()
            num_ntc = self.get1()
            d['temperatures'] = [self.get2() / 10 - 273.15 for _ in range(num_ntc)]
        elif self.address == 0x04:
            d['voltages'] = [self.get2() / 1000 for _ in range(self.length // 2)]
        elif self.address == 0x05:
            d['hardware version'] = "".join(map(chr, self.content))
        return d



from typing import Protocol, TypeVar

from .camera import CalibratableCamera

T = TypeVar('T', bound=CalibratableCamera)


class CalibratableCameraProvider(Protocol[T]):

    @property
    def cameras(self) -> dict[str, T]:
        ...



import random
from typing import Any, AsyncGenerator

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from .simulated_camera import SimulatedCamera


class SimulatedCameraProvider(CameraProvider[SimulatedCamera], persistence.PersistentModule):
    """This module collects and simulates cameras and generates synthetic images.

    In the current implementation the images only contain the camera ID and the current time.
    """
    SCAN_INTERVAL = 10

    def __init__(self, *,
                 simulate_failing: bool = False,
                 auto_scan: bool = True) -> None:
        super().__init__()

        self.simulate_device_failure = simulate_failing

        if auto_scan:
            rosys.on_repeat(self.update_device_list, self.SCAN_INTERVAL)

    def backup(self) -> dict:
        cameras = {}
        for camera in self._cameras.values():
            cameras[camera.id] = camera.to_dict()
        return {}

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            camera = SimulatedCamera.from_dict(camera_data)
            self.add_camera(camera)

        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    async def scan_for_cameras(self) -> AsyncGenerator[str, Any]:
        """Simulated device discovery by returning all camera's IDs.

        If simulate_device_failure is set, disconnected cameras are returned with a fixed probability.
        """
        for camera in self._cameras.values():
            if not camera.is_connected and self.simulate_device_failure:
                # return camera with fixed probability
                if random.random() < 0.5:
                    yield camera.id
            else:
                yield camera.id

    async def update_device_list(self) -> None:
        async for camera_id in self.scan_for_cameras():
            camera = self._cameras.get(camera_id)
            if not camera:
                camera = SimulatedCamera(id=camera_id, width=640, height=480)
                self.add_camera(camera)

            if not camera.is_connected:
                await camera.connect()
                continue
            assert camera.device is not None

            if self.simulate_device_failure:
                # disconnect cameras randomly with probability rising with time
                time_since_last_activation = rosys.time() - camera.device.creation_time
                if random.random() < time_since_last_activation / 30.0:
                    await camera.disconnect()

    def add_cameras(self, num_cameras: int) -> None:
        for _ in range(num_cameras):
            new_id = f'cam{len(self._cameras)}'
            self.add_camera(SimulatedCamera(id=new_id, width=640, height=480))



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Optional, overload

import cv2
import numpy as np

from ..geometry import Point, Point3d, Rotation
from .image import ImageSize


@dataclass(slots=True, kw_only=True)
class Intrinsics:
    matrix: list[list[float]]
    distortion: list[float]
    rotation: Rotation
    size: ImageSize

    @staticmethod
    def create_default(width: int = 800, height: int = 600, *, focal_length: float = 570) -> Intrinsics:
        size = ImageSize(width=width, height=height)
        K: list[list[float]] = [[focal_length, 0, size.width / 2], [0, focal_length, size.height / 2], [0, 0, 1]]
        D: list[float] = [0] * 5
        rotation = Rotation.zero()
        return Intrinsics(matrix=K, distortion=D, rotation=rotation, size=size)


log = logging.getLogger('rosys.world.calibration')


@dataclass(slots=True, kw_only=True)
class Extrinsics:
    rotation: Rotation = field(default_factory=lambda: Rotation.from_euler(np.pi, 0, 0))
    translation: list[float] = field(default_factory=lambda: [0.0, 0.0, 1.0])


@dataclass(slots=True, kw_only=True)
class Calibration:
    intrinsics: Intrinsics
    extrinsics: Extrinsics = field(default_factory=Extrinsics)

    @property
    def rotation(self) -> Rotation:
        return self.extrinsics.rotation * self.intrinsics.rotation

    @property
    def rotation_array(self) -> np.ndarray:
        return np.dot(self.extrinsics.rotation.R, self.intrinsics.rotation.R)

    @overload
    def project_to_image(self, world_coordinates: Point3d) -> Optional[Point]: ...

    @overload
    def project_to_image(self, world_coordinates: np.ndarray) -> np.ndarray: ...

    def project_to_image(self, world_coordinates: Point3d | np.ndarray) -> Optional[Point] | np.ndarray:
        if isinstance(world_coordinates, Point3d):
            world_array = np.array([world_coordinates.tuple], dtype=np.float32)
            image_array = self.project_to_image(world_array)
            if np.isnan(image_array).any():
                return None
            return Point(x=image_array[0, 0], y=image_array[0, 1])  # pylint: disable=unsubscriptable-object

        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion, dtype=float)
        image_array, _ = cv2.projectPoints(world_coordinates, Rod, t, K, D)
        local_coordinates = (world_coordinates - np.array(self.extrinsics.translation)) @ R
        image_array[local_coordinates[:, 2] < 0, :] = np.nan
        return image_array.reshape(-1, 2)

    @overload
    def project_from_image(self, image_coordinates: Point, target_height: float = 0) -> Optional[Point3d]: ...

    @overload
    def project_from_image(self, image_coordinates: np.ndarray, target_height: float = 0) -> np.ndarray: ...

    def project_from_image(self, image_coordinates: Point | np.ndarray, target_height: float = 0) -> Optional[Point3d] | np.ndarray:
        if isinstance(image_coordinates, Point):
            image_points = np.array(image_coordinates.tuple, dtype=np.float32)
            world_points = self.project_from_image(image_points, target_height=target_height)
            if np.isnan(world_points).any():
                return None
            return Point3d(x=world_points[0, 0], y=world_points[0, 1], z=world_points[0, 2])  # pylint: disable=unsubscriptable-object

        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_coordinates_ = cv2.undistortPoints(image_coordinates.astype(np.float32), K, D).reshape(-1, 2)
        image_coordinates__ = cv2.convertPointsToHomogeneous(image_coordinates_).reshape(-1, 3)
        objPoints = image_coordinates__ @ self.rotation_array.T
        Z = self.extrinsics.translation[-1]
        t = np.array(self.extrinsics.translation)
        world_points = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        reprojection = self.project_to_image(world_points)
        sign = objPoints[:, -1] * np.sign(Z)
        distance = np.linalg.norm(reprojection - image_coordinates, axis=1)
        world_points[np.logical_not(np.logical_and(sign < 0, distance < 2)), :] = np.nan

        return world_points

    @staticmethod
    def from_points(
            world_points: list[Point3d], image_points: list[Point], image_size: ImageSize, f0: float) -> Calibration:
        world_point_array = np.array([p.tuple for p in world_points], dtype=np.float32).reshape((1, -1, 3))
        image_point_array = np.array([p.tuple for p in image_points], dtype=np.float32).reshape((1, -1, 2))
        K0 = np.array([[f0, 0, image_size.width / 2], [0, f0, image_size.height / 2], [0, 0, 1]], dtype=np.float32)
        _, K, D, rvecs, tvecs = cv2.calibrateCamera(world_point_array, image_point_array, image_size.tuple, K0, None,
                                                    flags=cv2.CALIB_USE_INTRINSIC_GUESS)

        rotation0 = Rotation(R=np.eye(3).tolist())
        intrinsics = Intrinsics(matrix=K.tolist(), distortion=D.tolist()[0], rotation=rotation0, size=image_size)

        rotation = Rotation.from_rvec(rvecs[0]).T
        translation = (-np.array(rotation.R).dot(tvecs)).flatten().tolist()
        extrinsics = Extrinsics(rotation=rotation, translation=translation)

        return Calibration(intrinsics=intrinsics, extrinsics=extrinsics)



from .drivable import Drivable
from .driver import Driver, DrivingAbortedException
from .driver_object import DriverObject as driver_object
from .joystick_ import Joystick as joystick
from .keyboard_control_ import KeyboardControl as keyboard_control
from .odometer import Odometer, VelocityProvider
from .path_segment import PathSegment
from .robot_object_ import RobotObject as robot_object
from .steerer import Steerer

__all__ = [
    'Drivable',
    'Driver',
    'DrivingAbortedException',
    'driver_object',
    'joystick',
    'keyboard_control',
    'Odometer',
    'VelocityProvider',
    'PathSegment',
    'robot_object',
    'Steerer',
]



import abc

from ..event import Event
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class EStop(Module, abc.ABC):
    """A module that detects when the e-stop is triggered.

    The module has a boolean field `active` that is true when the e-stop is triggered.

    There is also a boolean field `is_soft_estop_active` that is true when the soft e-stop is active.
    It can be set to true or false by calling `set_soft_estop(active: bool)`.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.ESTOP_TRIGGERED = Event()
        """the e-stop was triggered"""

        self.active: bool = False
        self.is_soft_estop_active: bool = False
        self.pressed_estops: list[int] = []

    @abc.abstractmethod
    async def set_soft_estop(self, active: bool) -> None:
        if active:
            self.ESTOP_TRIGGERED.emit()
        self.is_soft_estop_active = active


class EStopHardware(EStop, ModuleHardware):
    """Hardware implementation of the e-stop module.

    The module expects a dictionary of pin names and pin numbers.
    """

    def __init__(self, robot_brain: RobotBrain, *, name: str = 'estop', pins: dict[str, int]) -> None:
        self.name = name
        self.pins = pins
        lizard_code = '\n'.join(f'{name}_{pin} = Input({number})' for pin, number in pins.items())
        core_message_fields = [f'{name}_{pin}.level' for pin in pins]
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code, core_message_fields=core_message_fields)

    async def set_soft_estop(self, active: bool) -> None:
        await super().set_soft_estop(active)
        await self.robot_brain.send(f'en3.level({"false" if active else "true"})')

    def handle_core_output(self, time: float, words: list[str]) -> None:
        corelist = [int(words.pop(0)) == 0 for _ in self.pins]
        active = any(corelist)
        pressed = [index for index, value in enumerate(corelist) if value]
        if pressed != self.pressed_estops:
            self.log.warning(f'E-Stop {pressed} changed')
        self.pressed_estops[:] = pressed
        if active and not self.active:
            self.ESTOP_TRIGGERED.emit()
        self.active = active


class EStopSimulation(EStop, ModuleSimulation):
    """Simulation of the e-stop module."""

    def activate(self) -> None:
        if not self.active:
            self.ESTOP_TRIGGERED.emit()
        self.active = True

    def deactivate(self) -> None:
        self.active = False

    async def set_soft_estop(self, active: bool) -> None:
        await super().set_soft_estop(active)



import io
import logging
from typing import Optional

import cv2
import imgsize
import numpy as np

from ..geometry import Rectangle
from .image import ImageSize
from .image_rotation import ImageRotation


class PeekableBytesIO(io.BytesIO):

    def peek(self, n=-1):
        position = self.tell()
        data = self.read(n)
        self.seek(position)
        return data


def get_image_size_from_bytes(image: bytes) -> ImageSize:
    try:
        with PeekableBytesIO(image) as f:
            width, height = imgsize.get_size(f)
    except imgsize.UnknownSize as e:
        raise ValueError('Could not determine image size') from e
    return ImageSize(width=width, height=height)


def encode_image_as_jpeg(image: np.ndarray) -> bytes:
    return cv2.imencode('.jpg', image)[1].tobytes()


def process_jpeg_image(data: bytes, rotation: ImageRotation, crop: Optional[Rectangle] = None) -> bytes | None:
    """Rotate and crop a JPEG image."""
    if crop is None and rotation == ImageRotation.NONE:
        return data
    array = np.frombuffer(data, dtype=np.uint8)
    decoded = cv2.imdecode(array, cv2.IMREAD_COLOR)
    if decoded is None:
        logging.warning('could not decode image buffer')
        return None
    return process_ndarray_image(decoded, rotation, crop)


def process_ndarray_image(image: np.ndarray, rotation: ImageRotation, crop: Optional[Rectangle] = None) -> bytes:
    """Rotate and crop a NumPy image and encode it as JPEG."""
    if crop is not None:
        image = image[int(crop.y):int(crop.y+crop.height), int(crop.x):int(crop.x+crop.width)]
    if rotation == ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    return encode_image_as_jpeg(image)


def remove_exif(image_data: bytes | bytearray) -> bytes:
    EXIF_MARKER = b'\xFF\xE1'

    pos = 2  # Skip SOI marker
    image_data = bytearray(image_data)
    while pos < len(image_data):
        match_start = image_data.find(EXIF_MARKER, pos)
        if match_start == -1:
            break
        match_end = match_start + 2
        length = int.from_bytes(image_data[match_end:match_end+2], byteorder='big')
        del image_data[match_start:match_end+length]  # Remove EXIF segment
        pos = match_end

    return bytes(image_data)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

import cv2
import numpy as np
from fastapi import Response
from nicegui import app

from .. import run
from .image import Image

if TYPE_CHECKING:
    from .camera import Camera

log = logging.getLogger('rosys.image_route')


def create_image_route(camera: Camera) -> None:
    placeholder_url = '/' + camera.base_path + '/placeholder'
    timestamp_url = '/' + camera.base_path + '/{timestamp}'

    app.remove_route(placeholder_url)
    app.remove_route(timestamp_url)

    async def get_camera_image(timestamp: str, shrink: int = 1) -> Response:
        return await _get_image(camera, timestamp, shrink)

    app.add_api_route(placeholder_url, _get_placeholder)
    app.add_api_route(timestamp_url, get_camera_image)


async def _get_placeholder(shrink: int = 1) -> Response:
    return Response(content=Image.create_placeholder('no image', shrink=shrink).data, media_type='image/jpeg')


async def _get_image(camera: Camera, timestamp: str, shrink: int = 1) -> Response:
    try:
        if not camera:
            return Response(content='Camera not found', status_code=404)
        jpeg = await _try_get_jpeg(camera, timestamp, shrink)
        if not jpeg:
            return Response(content='Image not found', status_code=404)
        return Response(content=jpeg, headers={'cache-control': 'max-age=7776000'}, media_type='image/jpeg')
    except Exception:
        log.exception('could not get image')
        raise


async def _try_get_jpeg(camera: Camera, timestamp: str, shrink: int) -> Optional[bytes]:
    for image in reversed(camera.images):
        if str(image.time) == timestamp and image.data is not None:
            jpeg = image.data
            if shrink != 1:
                array = np.frombuffer(image.data, dtype=np.uint8)
                if array is None:
                    return None
                jpeg = await run.cpu_bound(_shrink, shrink, array)
            return jpeg
    return None


def _shrink(factor: int, array: np.ndarray) -> bytes:
    decoded = cv2.imdecode(array, cv2.IMREAD_COLOR)
    img = decoded[::factor, ::factor]
    _, encoded_image = cv2.imencode('.jpg', img, [int(cv2.IMWRITE_JPEG_QUALITY), 60])
    return encoded_image.tobytes()



from dataclasses_json import Exclude, config

from .backup_schedule import BackupSchedule
from .converters import from_dict, replace_dataclass, replace_dict, replace_list, replace_set, to_dict
from .persistent_module import PersistentModule
from .registry import backup, restore, write_export
from .ui import export_button, import_button

exclude: dict[str, dict] = config(exclude=Exclude.ALWAYS)

__all__ = [
    'BackupSchedule',
    'from_dict',
    'replace_dataclass',
    'replace_dict',
    'replace_list',
    'replace_set',
    'to_dict',
    'PersistentModule',
    'exclude',
    'backup',
    'restore',
    'write_export',
    'export_button',
    'import_button',
]



from typing import Optional

from .axis_mjpeg_device import AxisMjpegDevice
from .mjpeg_device import MjpegDevice
from .motec_mjpeg_device import MotecMjpegDevice
from .vendors import VendorType, mac_to_vendor


class MjpegDeviceFactory:
    @staticmethod
    def create(mac: str, ip: str, *,
               index: Optional[int] = None,
               username: Optional[str] = None,
               password: Optional[str] = None,
               control_port: Optional[int] = None) -> MjpegDevice:

        if mac_to_vendor(mac) == VendorType.AXIS:
            return AxisMjpegDevice(mac, ip, index=index, username=username, password=password)

        if mac_to_vendor(mac) == VendorType.MOTEC:
            return MotecMjpegDevice(mac, ip, username=username, password=password, control_port=control_port)

        raise ValueError(f'Unknown vendor for mac="{mac}"')



from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True, kw_only=True)
class Prism:
    outline: list[tuple[float, float]]
    height: float

    @staticmethod
    def default_robot_shape() -> Prism:
        return Prism(outline=[(-0.5, -0.5), (0.5, -0.5), (0.75, 0), (0.5, 0.5), (-0.5, 0.5)], height=0.5)



from enum import Enum
from typing import Optional


class VendorType(Enum):
    JOVISION = 1
    DAHUA = 2
    AXIS = 3
    OTHER = -1


def mac_to_vendor(mac: str) -> VendorType:
    if mac.startswith('e0:62:90'):
        return VendorType.JOVISION
    if mac.startswith('e4:24:6c') or mac.startswith('3c:e3:6b'):
        return VendorType.DAHUA
    return VendorType.OTHER


def mac_to_url(mac: str, ip: str, jovision_profile: int = 0) -> Optional[str]:
    vendor = mac_to_vendor(mac)
    if vendor == VendorType.JOVISION:
        return f'rtsp://admin:admin@{ip}/profile{jovision_profile}'
    if vendor == VendorType.DAHUA:
        return f'rtsp://admin:Adminadmin@{ip}/cam/realmonitor?channel=1&subtype=0'
    return None



import logging
from typing import Any, Optional

from typing_extensions import Self

from ... import rosys
from ..camera import ConfigurableCamera, TransformableCamera
from ..image import Image
from ..image_processing import get_image_size_from_bytes, process_jpeg_image
from ..image_rotation import ImageRotation
from .mjpeg_device import MjpegDevice
from .mjpeg_device_factory import MjpegDeviceFactory


class MjpegCamera(TransformableCamera, ConfigurableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: str | None = None,
                 connect_after_init: bool = True,
                 streaming: bool = True,
                 polling_interval: float = 0.1,
                 base_path_overwrite: str | None = None,
                 username: str | None = None,
                 password: str | None = None,
                 ip: str | None = None,
                 **kwargs: Any,
                 ) -> None:
        super().__init__(id=id, name=name, connect_after_init=connect_after_init, streaming=streaming,
                         polling_interval=polling_interval, base_path_overwrite=base_path_overwrite, **kwargs)
        self.log = logging.getLogger(f'rosys.vision.mjpeg_camera.{self.id}')
        self.username = username
        self.password = password

        self.ip = ip

        self.index: Optional[int] = None
        parts = self.id.split('-')
        if len(parts) == 2 and parts[1].isdigit():
            self.index = int(parts[1])

        self.mac = parts[0]
        self.device: Optional[MjpegDevice] = None

        self._register_parameter('fps', self._get_fps, self._set_fps, default_value=10)
        self._register_parameter('resolution', self._get_resolution, self._set_resolution, default_value=(640, 480))
        self._register_parameter('mirrored', self._get_mirrored, self._set_mirrored, default_value=False)

    def to_dict(self) -> dict:
        return super().to_dict() | {
            'username': self.username,
            'password': self.password,
            'ip': self.ip,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(**data)

    @property
    def is_connected(self) -> bool:
        return (self.device is not None) and (self.device.capture_task is not None) and (not self.device.capture_task.done())

    async def connect(self) -> None:
        if self.is_connected:
            return

        if not self.ip:
            self.log.error('No IP address provided')
            return

        try:
            self.device = MjpegDeviceFactory.create(self.mac, self.ip, index=self.index, username=self.username,
                                                    password=self.password)
        except ValueError as error:
            self.log.error('Could not connect to device: %s', error)
            return

    async def disconnect(self) -> None:
        if self.device is None:
            return

        self.device.shutdown()
        self.device = None

    async def capture_image(self) -> None:
        if not self.is_connected:
            return

        assert self.device is not None

        image = self.device.capture()
        if image is None:
            return

        if self.crop or self.rotation != ImageRotation.NONE:
            image = await rosys.run.cpu_bound(process_jpeg_image, image, self.rotation, self.crop)
        if image is None:
            return
        try:
            final_image_resolution = get_image_size_from_bytes(image)
        except ValueError:
            return

        self._add_image(Image(camera_id=self.id, data=image, time=rosys.time(), size=final_image_resolution))

    async def _set_fps(self, fps: int) -> None:
        if self.device is None:
            raise ValueError('Device is not connected')

        await self.device.set_fps(fps)
        self.polling_interval = 1.0 / fps

    async def _get_fps(self) -> int:
        if self.device is None:
            raise ValueError('Device is not connected')

        return await self.device.get_fps()

    async def _set_resolution(self, resolution: tuple[int, int]) -> None:
        if self.device is None:
            raise ValueError('Device is not connected')

        await self.device.set_resolution(*resolution)

    async def _get_resolution(self) -> tuple[int, int]:
        if self.device is None:
            raise ValueError('Device is not connected')

        return await self.device.get_resolution()

    async def _set_mirrored(self, mirrored: bool) -> None:
        if self.device is None:
            raise ValueError('Device is not connected')

        await self.device.set_mirrored(mirrored)

    async def _get_mirrored(self) -> bool:
        if self.device is None:
            raise ValueError('Device is not connected')

        return await self.device.get_mirrored()



import asyncio
from typing import Any, Coroutine, TypeVar

_T = TypeVar('_T')


class LazyWorker:

    def __init__(self) -> None:
        self._condition = asyncio.Event()
        self._is_free = True

    async def run(self, coro: Coroutine[Any, None, _T]) -> _T | None:
        self._notify()

        if not self._is_free:
            await self._condition.wait()

        if not self._is_free:
            coro.close()
            return None

        self._is_free = False
        result = await coro
        self._is_free = True
        self._notify()
        return result

    def _notify(self) -> None:
        self._condition.set()
        self._condition.clear()



import re
from dataclasses import dataclass
from typing import Optional

from .mjpeg_device import MjpegDevice
from .vendors import VendorType, mac_to_vendor


@dataclass(slots=True, kw_only=True)
class AxisSettings:
    fps: int
    resolution: tuple[int, int]
    mirrored: bool

    def __post_init__(self):
        if self.fps <= 0:
            raise ValueError(f'fps must be positive. Got {self.fps}')
        if self.resolution[0] <= 0 or self.resolution[1] <= 0:
            raise ValueError(f'resolution must be positive. Got {self.resolution}')
        if not isinstance(self.mirrored, bool):
            raise ValueError(f'mirrored must be a boolean. Got {self.mirrored}')


class AxisMjpegDevice(MjpegDevice):
    def __init__(self, mac: str, ip: str, *,
                 index: Optional[int] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None) -> None:
        super().__init__(mac, ip,
                         index=index,
                         username=username, password=password)

        self.axis_settings = AxisSettings(fps=6, resolution=(640, 480), mirrored=False)

        vendor = mac_to_vendor(mac)
        if vendor != VendorType.AXIS:
            raise ValueError(f'AxisMjpegDevice can only be used with AXIS devices. Got {vendor} for mac="{mac}"')

    async def get_fps(self) -> int:
        return self.axis_settings.fps

    async def set_fps(self, fps: int) -> None:
        self.url = re.sub(r'fps=\d+', f'fps={fps}', self.url)
        if 'fps=' not in self.url:
            self.url += f'&fps={fps}'
        self.restart_capture()

    async def get_resolution(self) -> tuple[int, int]:
        return self.axis_settings.resolution

    async def set_resolution(self, width: int, height: int) -> None:
        self.url = re.sub(r'resolution=\d+x\d+', f'resolution={width}x{height}', self.url)
        if 'resolution=' not in self.url:
            self.url += f'&resolution={width}x{height}'
        self.restart_capture()

    async def get_mirrored(self) -> bool:
        return self.axis_settings.mirrored

    async def set_mirrored(self, mirrored: bool) -> None:
        value = 1 if mirrored else 0
        self.url = re.sub(r'mirror=\d', f'mirror={value}', self.url)
        if 'mirror=' not in self.url:
            self.url += f'&mirror={value}'
        self.restart_capture()



import logging
from typing import Optional

import httpx

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from ..rtsp_camera.arp_scan import find_cameras
from .mjpeg_camera import MjpegCamera
from .vendors import VendorType, mac_to_vendor


class MjpegCameraProvider(CameraProvider[MjpegCamera], persistence.PersistentModule):
    SCAN_INTERVAL = 10

    def __init__(self, *,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 network_interface: Optional[str] = None,
                 auto_scan: bool = True) -> None:
        """CameraProvider for MJpegCamera

        :param username: username to assign for new cameras
        :param password: password to assign for new cameras
        :param network_interface: network interface used to scan for cameras
        """
        super().__init__()

        self.username = username
        self.password = password
        self.network_interface = network_interface

        self.log = logging.getLogger('rosys.mjpeg_camera_provider')
        rosys.on_shutdown(self.shutdown)
        if auto_scan:
            rosys.on_repeat(self.update_device_list, self.SCAN_INTERVAL)

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            self.log.debug('restoring camera: %s', camera_data)
            camera = MjpegCamera.from_dict(camera_data)
            self.add_camera(camera)
        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    async def scan_for_cameras(self) -> list[tuple[str, str]]:
        self.log.debug('scanning for cameras...')
        ids_ips: list[tuple[str, str]] = []
        async for mac, ip in find_cameras(self.network_interface):
            self.log.debug('device found: %s %s', mac, ip)
            vendor = mac_to_vendor(mac)
            if vendor == VendorType.OTHER:
                continue

            if vendor == VendorType.AXIS:
                authentication = None if self.username is None or self.password is None else \
                    httpx.DigestAuth(self.username, self.password)
                url = f'http://{ip}/axis-cgi/videostatus.cgi'
                try:
                    async with httpx.AsyncClient() as client:
                        response = await client.get(url, auth=authentication)
                except httpx.HTTPError as e:
                    self.log.warning('Error while looking for cameras at axis router (%s): %s', url, e)
                    continue
                if response.status_code != 200:
                    self.log.warning('Error while looking for cameras at axis router (%s): %s', url, response.text)
                    continue

                self.log.debug('found axis camera at ip %s', ip)
                self.log.debug('video status:\n"%s\n"', response.text)
                camera_infos = response.text.split('\n')
                for info_line in camera_infos:
                    if info_line.endswith(' = video'):
                        video_name = info_line.removesuffix(' = video')
                        index = video_name.split(' ')[1]
                        ids_ips.append((f'{mac}-{index}', ip))
            else:
                ids_ips.append((mac, ip))
        self.log.debug('scanning done. Found %d cameras', len(ids_ips))
        return ids_ips

    async def update_device_list(self) -> None:
        for camera_id, ip in await self.scan_for_cameras():
            if camera_id not in self._cameras:
                self.log.info('found new camera "%s" at ip "%s"', camera_id, ip)
                self.add_camera(MjpegCamera(id=camera_id, username=self.username,
                                password=self.password, ip=ip))
            camera = self._cameras[camera_id]
            if not camera.is_connected:
                self.log.info('activating camera "%s" at ip "%s" ...', camera.id, ip)
                camera.ip = ip
                await camera.connect()

    async def shutdown(self) -> None:
        for camera in self._cameras.values():
            await camera.disconnect()



import functools
import logging
from datetime import datetime, timedelta
from typing import Any, Callable, Optional, cast

import suntime
from nicegui import ui

from .. import persistence, rosys
from ..event import Event
from .automator import Automator


class Translator:
    Location = 'Location'
    Sunrise_offset = 'Sunrise offset'
    Sunset_offset = 'Sunset offset'
    min = 'min'
    enabled = 'enabled'
    Mon = 'Mon'
    Tue = 'Tue'
    Wed = 'Wed'
    Thu = 'Thu'
    Fri = 'Fri'
    Sat = 'Sat'
    Sun = 'Sun'


class Schedule(persistence.PersistentModule):

    def __init__(self, *,
                 automator: Optional[Automator] = None,
                 on_activate: Optional[Callable] = None,
                 on_deactivate: Optional[Callable] = None,
                 location: Optional[tuple[float, float]] = None,
                 locations: Optional[dict[tuple[float, float], str]] = None,
                 sunrise_offset: float = 0.0,
                 sunset_offset: float = 0.0,
                 is_enabled: bool = False,
                 ) -> None:
        """Schedules automations according to a time plan.

        :param automator: optional automator to use for executing the automations
        :param on_activate: automation to execute when entering an active period
        :param on_deactivate: automation to execute when entering an inactive period
        :param location: optional geographic coordinates to activate the robot at daylight only
        :param locations: optional dictionary with geographic coordinates to choose from
        :param sunrise_offset: optional offset for activation after sunrise (minutes, default: 0.0)
        :param sunset_offset: optional offset for deactivation after sunset (minutes, default: 0.0)
        :param is_enabled: whether the schedule is enabled (default: False)
        """
        super().__init__()

        self.log = logging.getLogger('rosys.schedule')

        self.SCHEDULE_CHANGED = Event()
        """the schedule has changed"""

        self.automator = automator
        self.translator: Translator = rosys.translator or Translator()
        self.on_activate = on_activate
        self.on_deactivate = on_deactivate
        self.location = location
        self.locations = locations
        self.sunrise_offset = sunrise_offset
        self.sunset_offset = sunset_offset
        self.is_enabled = is_enabled

        self.half_hours = [True] * 2 * 24 * 7

        self._is_active = False
        rosys.on_repeat(self.step, 1)

    def backup(self) -> dict:
        return {
            'location': self.location,
            'sunrise_offset': self.sunrise_offset,
            'sunset_offset': self.sunset_offset,
            'is_enabled': self.is_enabled,
            'is_active': self._is_active,
            'half_hours': self.half_hours,
        }

    def restore(self, data: dict[str, Any]) -> None:
        self.location = (data['location'][0], data['location'][1]) if data.get('location') else None
        self.sunrise_offset = cast(float, data.get('sunrise_offset'))
        self.sunset_offset = cast(float, data.get('sunset_offset'))
        self.is_enabled = cast(bool, data.get('is_enabled', False))
        self._is_active = cast(bool, data.get('is_active', False))
        self.half_hours[:] = cast(list[bool], data['half_hours']) if 'half_hours' in data else [True] * 2 * 24 * 7
        self.SCHEDULE_CHANGED.emit()

    def invalidate(self) -> None:
        self.request_backup()
        self.SCHEDULE_CHANGED.emit()

    def time_to_index(self, weekday: int, hour: int, minute: int) -> int:
        if not 0 <= weekday <= 6:
            raise ValueError(f'day of week must be between 0 and 6, not {weekday}')
        if not 0 <= minute <= 60:
            raise ValueError(f'minutes must be between 0 and 60, not {minute}')
        return weekday * 2 * 24 + hour * 2 + minute // 30

    def is_dark(self, hour: int, minute: int) -> bool:
        sun_start_hour = _get_sun_start_hour(self.location, self.sunrise_offset)
        sun_stop_hour = _get_sun_stop_hour(self.location, self.sunset_offset)
        return not sun_start_hour < hour + minute / 60 < sun_stop_hour

    def is_planned(self, weekday: int, hour: int, minute: Optional[int] = None) -> bool:
        if minute is None:
            return self.half_hours[self.time_to_index(weekday, hour, 0)] \
                or self.half_hours[self.time_to_index(weekday, hour, 30)]
        return self.half_hours[self.time_to_index(weekday, hour, minute)]

    def fill(self, value: bool) -> None:
        for i, _ in enumerate(self.half_hours):
            self.half_hours[i] = value

    def can_be_active(self) -> bool:
        now = datetime.fromtimestamp(rosys.time())
        return self.is_planned(now.weekday(), now.hour, now.minute) and not self.is_dark(now.hour, now.minute)

    def step(self) -> None:
        if not self.is_enabled:
            return
        if not self._is_active and self.can_be_active():
            self.log.info('activate')
            if self.on_activate:
                self.log.info('start automation')
                if self.automator:
                    self.automator.start(self.on_activate(), paused=self.automator.is_paused)
                else:
                    self.on_activate()
            self._is_active = True
        if self._is_active and not self.can_be_active():
            self.log.info('deactivate')
            if self.on_deactivate:
                self.log.info('start automation')
                if self.automator:
                    self.automator.start(self.on_deactivate(), paused=self.automator.is_paused)
                else:
                    self.on_deactivate()
            self._is_active = False

    def ui(self) -> ui.column:
        def update() -> None:
            def color(weekday: int, hour: int, minutes: list[int]) -> str:
                # https://maketintsandshades.com/#21ba45,c10015
                if all(self.is_dark(hour, minute) for minute in minutes):
                    return '#d3f1da' if self.is_planned(weekday, hour, minutes[0]) else '#f3ccd0'
                dt = datetime.fromtimestamp(rosys.time())
                is_now = dt.weekday() == weekday and dt.hour == hour
                positive = '#147029' if is_now else '#21ba45' if d < 5 else '#1a9537'
                negative = '#74000d' if is_now else '#c10015' if d < 5 else '#9a0011'
                return positive if self.is_planned(weekday, hour, minutes[0]) else negative
            if not buttons:
                return
            for d in range(7):
                for h in range(24):
                    b0, b1, b2 = buttons[d * 24 + h]
                    b0.classes(replace=f'!bg-[{color(d, h, [0, 30, 60])}]')
                    b1.classes(replace=f'!bg-[{color(d, h, [0, 30])}]')
                    b2.classes(replace=f'!bg-[{color(d, h, [30, 60])}]')

        t = self.translator
        buttons: list[tuple[ui.button, ui.button, ui.button]] = []
        with ui.column() as grid:
            with ui.row().classes('fit items-center justify-between'):
                ui.switch(t.enabled).bind_value(self, 'is_enabled')
                if self.locations:
                    ui.select(self.locations, label=t.Location, on_change=lambda e: self.set_location(e.value)) \
                        .bind_value(self, 'location').style('width: 21em')
                    ui.number(t.Sunrise_offset, format='%.0f', on_change=self.invalidate, suffix=t.min) \
                        .bind_value(self, 'sunrise_offset').style('width:100px')
                    ui.number(t.Sunset_offset, format='%.0f', on_change=self.invalidate, suffix=t.min) \
                        .bind_value(self, 'sunset_offset').style('width:100px')
            with ui.column().style('gap: 0.3em'):
                for d, day in enumerate([t.Mon, t.Tue, t.Wed, t.Thu, t.Fri, t.Sat, t.Sun]):
                    with ui.row().style('gap: 0.3em'):
                        ui.label(day).classes('mt-2').style('width: 2em')
                        for h in range(24):
                            with ui.column().style('gap: 0.1em'):
                                hour = ui.button(f'{h:02d}', on_click=lambda _, d=d, h=h: self._toggle(d, h)) \
                                    .props('unelevated dense')
                                with ui.row().style('gap: 0.1em'):
                                    first_half = ui.button(on_click=lambda _, d=d, h=h: self._toggle(d, h, 0)) \
                                        .props('unelevated dense').style('height: 1em; width: 0.8em')
                                    second_half = ui.button(on_click=lambda _, d=d, h=h: self._toggle(d, h, 30)) \
                                        .props('unelevated dense').style('height: 1em; width: 0.8em')
                            buttons.append((hour, first_half, second_half))

        ui.timer(60, update)  # NOTE: to update the "now" indicator
        self.SCHEDULE_CHANGED.register_ui(update)
        return grid

    def set_location(self, location: tuple[float, float]) -> None:
        self.location = location
        self.invalidate()

    def _toggle(self, weekday: int, hour: int, minute: Optional[int] = None) -> None:
        new_value = not self.is_planned(weekday, hour, minute)
        if minute is None:
            self.half_hours[self.time_to_index(weekday, hour, 0)] = new_value
            self.half_hours[self.time_to_index(weekday, hour, 30)] = new_value
        else:
            self.half_hours[self.time_to_index(weekday, hour, minute)] = new_value
        self.invalidate()


@functools.lru_cache(maxsize=100)
def _get_sun_start_hour(location: tuple[float, float], offset: float) -> float:
    if not location:
        return 0.0
    sun = suntime.Sun(lat=location[0], lon=location[1])
    start_time = sun.get_sunrise_time(datetime.now()).astimezone() + timedelta(minutes=offset)
    return start_time.hour + start_time.minute / 60 + start_time.second / 60 / 60


@functools.lru_cache(maxsize=100)
def _get_sun_stop_hour(location, offset) -> float:
    if not location:
        return 24.0
    sun = suntime.Sun(lat=location[0], lon=location[1])
    stop_time = sun.get_sunset_time(datetime.now()).astimezone() + timedelta(minutes=offset)
    return stop_time.hour + stop_time.minute / 60 + stop_time.second / 60 / 60



import os
import sys
from typing import AsyncIterator, Optional

import ifaddr

from ... import rosys
from .vendors import VendorType, mac_to_vendor


def get_network_interface() -> str | None:
    """Return the first network interface that is not a loopback or virtual interface."""
    for interface in ifaddr.get_adapters():
        if any(interface.name.startswith(prefix) for prefix in ['lo', 'can', 'docker', 'veth', 'br']):
            continue
        if interface.ips is None:
            continue
        return interface.name
    return None


async def run_arp_scan(interface: Optional[str] = None) -> str:
    """Run an arp-scan on the given interface and return the output.
    :param interface: The network interface to use for the arp-scan.
    """
    if sys.platform.startswith('darwin'):
        arpscan_cmd = 'arp-scan'
    else:
        arpscan_cmd = '/usr/sbin/arp-scan'  # TODO is this necessary? sbin should be in path
    if os.getuid() != 0:
        arpscan_cmd = f'sudo {arpscan_cmd}'

    cmd = f'{arpscan_cmd} --localnet'
    if not interface:
        interface = get_network_interface()
    if interface:
        cmd += f' --interface {interface}'

    output = await rosys.run.sh(cmd, timeout=10)

    if 'sudo' in output and 'name resolution' not in output:
        raise RuntimeError('Could not run arp-scan! Make sure it is installed and can be run with sudo.'
                           'Try running sudo visudo and add the following line: "rosys ALL=(ALL) NOPASSWD: /usr/sbin/arp-scan"')
    if 'Could not obtain MAC address for interface' in output:
        raise RuntimeError(f'arp-scan failed to obtain MAC address for interface {interface}')
    return output


async def find_cameras(network_interface: Optional[str] = None) -> AsyncIterator[tuple[str, str]]:
    """Find all cameras in the local network and return MAC and IP addresses.
    :param network_interface: The network interface to use for the arp-scan.
    """
    output = await run_arp_scan(interface=network_interface)
    if 'ERROR' in output:
        return
    for line in output.splitlines():
        infos = line.split()
        if len(infos) < 2:
            continue
        mac = infos[1]
        ip = infos[0]
        if mac in ['packets', 'arp-scan'] or ip == 'Interface:':
            continue
        yield mac, ip


async def find_known_cameras(network_interface: Optional[str] = None) -> list[tuple[str, str]]:
    """Find all cameras of known vendors and return MAC addresses.
    :param network_interface: The network interface to use for the arp-scan.
    """
    return [(mac, ip) async for mac, ip in find_cameras(network_interface=network_interface) if mac_to_vendor(mac) != VendorType.OTHER]


async def find_ip(mac: str, network_interface: Optional[str] = None) -> Optional[str]:
    """Find the IP address of a camera with the given MAC address."""
    async for mac_, ip in find_cameras(network_interface=network_interface):
        if mac_ == mac:
            return ip
    return None



import logging
from typing import Optional

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from .arp_scan import find_known_cameras
from .rtsp_camera import RtspCamera


class RtspCameraProvider(CameraProvider[RtspCamera], persistence.PersistentModule):
    """This module collects and provides real RTSP streaming cameras."""
    SCAN_INTERVAL = 10

    def __init__(self, *,
                 frame_rate: int = 6,
                 jovision_profile: int = 0,
                 network_interface: Optional[str] = None,
                 auto_scan: bool = True) -> None:
        super().__init__()

        self.frame_rate = frame_rate
        self.jovision_profile = jovision_profile
        self.network_interface = network_interface

        self.log = logging.getLogger('rosys.rtsp_camera_provider')

        rosys.on_shutdown(self.shutdown)
        if auto_scan:
            rosys.on_repeat(self.update_device_list, self.SCAN_INTERVAL)

    def backup(self) -> dict:
        cameras = {}
        for camera in self._cameras.values():
            cameras[camera.id] = camera.to_dict()
        return {'cameras': cameras}

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            camera = RtspCamera.from_dict(camera_data)
            self.add_camera(camera)
        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    @staticmethod
    async def scan_for_cameras(network_interface: Optional[str] = None) -> list[tuple[str, str]]:
        return await find_known_cameras(network_interface=network_interface)

    async def update_device_list(self) -> None:
        for mac, ip in await find_known_cameras(network_interface=self.network_interface):
            if mac not in self._cameras:
                self.add_camera(RtspCamera(id=mac, fps=self.frame_rate, jovision_profile=self.jovision_profile, ip=ip))
            camera = self._cameras[mac]
            if not camera.is_connected:
                self.log.info('activating authorized camera %s...', camera.id)
                camera.ip = ip
                await camera.connect()

    async def shutdown(self) -> None:
        for camera in self._cameras.values():
            await camera.disconnect()



from __future__ import annotations

import io
import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Optional, Protocol

import humanize
from PIL import Image, ImageDraw, ImageFont

from .. import rosys
from ..vision import Camera

STORAGE_PATH = Path('~/.rosys/timelapse').expanduser()
VIDEO_PATH = STORAGE_PATH / 'videos'
FONT = str(Path(__file__).parent / 'assets' / 'RobotoMono-Medium.ttf')
IMAGE_FONT = ImageFont.truetype(FONT, 16)
BIG_COVER_FONT = ImageFont.truetype(FONT, 50)
SMALL_COVER_FONT = ImageFont.truetype(FONT, 30)


class RosysImage(Protocol):
    camera_id: str
    time: float
    data: Optional[bytes] = None


class TimelapseRecorder:

    def __init__(self, *, width: int = 800, height: int = 600, capture_rate: float = 1) -> None:
        """Creates a time lapse recorder to capture images from a camera and creates a video of the sequence afterwards.

        param: width: width of the images to capture (default: 800)
        param: height: height of the images to capture (default: 600)
        param: capture_rate: images per second to capture (default: 1)
        """
        self.log = logging.getLogger('rosys.timelapse_recorder')
        self.width = width
        self.height = height
        self.avoid_broken_images = True
        self.capture_rate = capture_rate
        self.last_capture_time = rosys.time()
        self._notifications: list[list[str]] = []
        self.camera: Optional[Camera] = None
        VIDEO_PATH.mkdir(parents=True, exist_ok=True)
        rosys.on_repeat(self._capture, 0.01)

    async def _capture(self) -> None:
        if self.camera is None:
            return
        if rosys.time() - self.last_capture_time < 1 / self.capture_rate:
            return
        images = self.camera.get_recent_images(timespan=2)
        if self.avoid_broken_images:
            images = [i for i in images if not i.is_broken and i.time < rosys.time() - 0.1]
        if images:
            self.last_capture_time = rosys.time()
            await self.save(images[-1])

    async def save(self, image: RosysImage) -> None:
        """Captures an image to be used in video."""
        await rosys.run.cpu_bound(_save_image,
                                  image,
                                  STORAGE_PATH,
                                  (self.width, self.height),
                                  self._notifications.pop(0) if self._notifications else [])

    async def compress_video(self) -> None:
        """Creates a video from the captured images"""
        jpgs = sorted(STORAGE_PATH.glob('*.jpg'))
        if len(jpgs) < 20:
            self.log.info('very few images (%s); not creating video', len(jpgs))
            self.discard_video()
            return
        start = datetime.fromtimestamp(float(jpgs[0].stem))
        end = datetime.fromtimestamp(float(jpgs[-1].stem))
        self.log.info('creating video from %s to %s', start, end)
        duration = humanize.naturaldelta(end - start)
        await self.create_info(start.strftime(r'%d.%m.%Y %H:%M:%S'), duration, time=start.timestamp() - 1)
        id_ = start.strftime(r'%Y%m%d_%H-%M-%S_' + duration.replace(' ', '_'))
        target_dir = STORAGE_PATH / id_
        target_dir.mkdir(parents=True, exist_ok=True)
        await rosys.run.sh(f'mv {STORAGE_PATH}/*.jpg {target_dir}', shell=True)
        source_file = target_dir / 'source.txt'
        with source_file.open('w') as f:
            for jpg in sorted(target_dir.glob('*.jpg')):
                f.write(f"file '{jpg}'\n")
        absolute_niceness = 10 - os.nice(0)
        cmd = (
            f'nice -n {absolute_niceness} ffmpeg -hide_banner -threads 1 -f concat -safe 0 -i "{source_file}" '
            f'-s {self.width}x{self.height} -vcodec libx264 -crf 18 -preset slow -pix_fmt yuv420p -y {target_dir}/{id_}.mp4;'
            f'mv {target_dir}/*mp4 {STORAGE_PATH}/videos;'
            f'rm -r {target_dir};'
        )
        self.log.info('starting %s', cmd)
        rosys.background_tasks.create(rosys.run.sh(cmd, timeout=None, shell=True), name='timelapse ffmpeg')

    def discard_video(self) -> None:
        """Drop the currently recorded video data."""
        for jpg in STORAGE_PATH.glob('*.jpg'):
            jpg.unlink()

    async def create_info(self, title: str, subtitle: str, frames: int = 20, time: Optional[float] = None) -> None:
        """Shows info on black screen with title and subtitle"""
        await rosys.run.cpu_bound(_save_info, title, subtitle, rosys.time() if time is None else time, frames)

    def notify(self, message: str, frames: int = 10) -> None:
        """Place message in the next n frames"""
        while len(self._notifications) < frames:
            self._notifications.append([])
        for i in range(frames):
            self._notifications[i].append(message)


def _save_image(image: RosysImage, path: Path, size: tuple[int, int], notifications: list[str]) -> None:
    assert image.data is not None
    img = Image.open(io.BytesIO(image.data))
    img = img.resize(size)
    draw = ImageDraw.Draw(img)
    x = y = 20
    _write(f'{datetime.fromtimestamp(image.time):%Y-%m-%d %H:%M:%S}, cam {image.camera_id}', draw, x, y)
    for message in notifications:
        y += 30
        _write(message, draw, x, y)
    img.save(path / f'{image.time:.3f}.jpg', 'JPEG')


def _write(text: str, draw: ImageDraw.ImageDraw, x: int, y: int) -> None:
    # shadow
    draw.text((x - 1, y - 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x + 1, y - 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x - 1, y + 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x + 1, y + 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    # actual text
    draw.text((x, y), text, font=IMAGE_FONT, fill=(255, 255, 255))


def _save_info(title: str, subtitle: str, time: float, frames: int) -> None:
    for i in range(frames):
        img = Image.new('RGB', (1600, 1200), 'black')
        draw = ImageDraw.Draw(img)
        font = BIG_COVER_FONT if len(subtitle) < 25 and len(title) < 25 else SMALL_COVER_FONT

        title_box = draw.textbbox((0, 0), title, font=font)
        title_x = (img.width - title_box[2]) / 2
        title_y = (img.height / 2) - 200 - (title_box[3] / 2)
        draw.text((title_x, title_y), title, font=font, fill='white')

        subtitle_box = draw.textbbox((0, 0), subtitle, font=font)
        subtitle_x = (img.width - subtitle_box[2]) / 2
        subtitle_y = (img.height / 2) + 100 - (subtitle_box[3] / 2)
        draw.text((subtitle_x, subtitle_y), subtitle, font=font, fill='white')

        img.save(STORAGE_PATH / f'{time:.3f}{i:03}.jpg')



import numpy as np
from matplotlib.path import Path


class BinaryRenderer:

    def __init__(self, size, fill_value: bool = False) -> None:
        self.map = np.full(size, fill_value=fill_value, dtype=bool)

        self.xx, self.yy = np.meshgrid(range(size[1]), range(size[0]))
        self.xy = np.vstack((self.xx.flatten(), self.yy.flatten())).T

    def circle(self, x, y, radius, value=True) -> None:
        x0 = max(int(x - radius), 0)
        y0 = max(int(y - radius), 0)
        x1 = min(int(x + radius) + 2, self.map.shape[1] - 1)
        y1 = min(int(y + radius) + 2, self.map.shape[0] - 1)
        roi = self.map[y0:y1, x0:x1]
        sqr_dist = (self.xx[y0:y1, x0:x1] - x)**2 + (self.yy[y0:y1, x0:x1] - y)**2
        roi[sqr_dist <= radius**2] = value
        self.map[y0:y1, x0:x1] = roi

    def polygon(self, points, value=True) -> None:
        if len(points) == 0:
            return
        x0 = max(int(points[:, 0].min()), 0)
        y0 = max(int(points[:, 1].min()), 0)
        x1 = min(int(points[:, 0].max()) + 2, self.map.shape[1] - 1)
        y1 = min(int(points[:, 1].max()) + 2, self.map.shape[0] - 1)
        xy = np.vstack((self.xx[y0:y1, x0:x1].flatten(), self.yy[y0:y1, x0:x1].flatten())).T
        roi = self.map[y0:y1, x0:x1]
        roi[Path(points).contains_points(xy).reshape(roi.shape)] = value
        self.map[y0:y1, x0:x1] = roi



from .calibratable_camera_provider import CalibratableCameraProvider
from .calibration import Calibration, Extrinsics, Intrinsics
from .camera import CalibratableCamera, Camera, ConfigurableCamera, TransformableCamera
from .camera_objects_ import CameraObjects as camera_objects
from .camera_projector import CameraProjector
from .camera_provider import CameraProvider
from .detections import BoxDetection, Detection, Detections, PointDetection
from .detector import Autoupload, Detector
from .detector_hardware import DetectorHardware
from .detector_simulation import DetectorSimulation, SimulatedObject
from .image import Image, ImageSize
from .mjpeg_camera import MjpegCamera, MjpegCameraProvider
from .multi_camera_provider import MultiCameraProvider
from .rtsp_camera import RtspCamera, RtspCameraProvider
from .simulated_camera import SimulatedCamera, SimulatedCameraProvider
from .usb_camera import UsbCamera, UsbCameraProvider

__all__ = [
    'CalibratableCameraProvider',
    'Calibration',
    'Extrinsics',
    'Intrinsics',
    'CalibratableCamera',
    'Camera',
    'ConfigurableCamera',
    'TransformableCamera',
    'camera_objects',
    'CameraProjector',
    'CameraProvider',
    'BoxDetection',
    'Detection',
    'Detections',
    'PointDetection',
    'Autoupload',
    'Detector',
    'DetectorHardware',
    'DetectorSimulation',
    'SimulatedObject',
    'Image',
    'ImageSize',
    'MjpegCamera',
    'MjpegCameraProvider',
    'MultiCameraProvider',
    'RtspCamera',
    'RtspCameraProvider',
    'SimulatedCamera',
    'SimulatedCameraProvider',
    'UsbCamera',
    'UsbCameraProvider',
]



from typing import Coroutine


class parallelize:
    """Parallelize multiple coroutines.

    This class allows to combine multiple coroutines into one that can be passed to the
    `automator <https://rosys.io/reference/rosys/automation/#rosys.automation.Automator>`__
    to run them in parallel.
    """

    def __init__(self, *coros: Coroutine, return_when_first_completed: bool = False) -> None:
        self.coros = coros
        self.return_when_first_completed = return_when_first_completed

    def __await__(self):
        coro_iters = [coro.__await__() for coro in self.coros]
        iter_sends = [coro_iter.send for coro_iter in coro_iters]
        iter_throws = [coro_iter.throw for coro_iter in coro_iters]
        sends = iter_sends[:]
        messages = [None for _ in sends]
        completed = [False for _ in sends]
        while not all(completed):
            for i, send in enumerate(sends):
                if completed[i]:
                    continue
                try:
                    signal = send(messages[i])
                except StopIteration:
                    if self.return_when_first_completed:
                        return
                    completed[i] = True
                    continue
                else:
                    sends[i] = iter_sends[i]
                try:
                    messages[i] = yield signal
                except GeneratorExit:
                    raise
                except BaseException as e:
                    sends[i], messages[i] = iter_throws[i], e



from dataclasses import dataclass
from itertools import combinations, pairwise
from typing import Optional

from typing_extensions import Self

from ..geometry import LineSegment, Point, Polygon


@dataclass(slots=True, kw_only=True)
class Area(Polygon):
    id: str
    type: Optional[str] = None
    color: str = 'green'
    closed: bool = True

    def close(self) -> Self:
        self.closed = True
        return self

    def would_cause_self_intersection(self, point: Point, new_index: Optional[int] = None) -> bool:
        new_outline = self.outline[:]
        if new_index is None:
            new_outline.append(point)
        else:
            new_outline.insert(new_index, point)
        line_segments = [LineSegment(point1=p1, point2=p2) for p1, p2 in pairwise(new_outline + new_outline[:1])]
        if not self.closed:
            line_segments.pop()
        for segment1, segment2 in combinations(line_segments, 2):
            if id(point) not in {id(segment1.point1), id(segment1.point2), id(segment2.point1), id(segment2.point2)}:
                continue  # segments are not affected by the new point
            if {id(segment1.point1), id(segment1.point2)}.intersection({id(segment2.point1), id(segment2.point2)}):
                continue  # segments are connected by a shared point
            if segment1.intersect(segment2):
                return True
        return False



import abc
import logging
from typing import Optional, cast

from .. import rosys
from ..helpers import remove_indentation
from .expander import ExpanderHardware
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class Robot(abc.ABC):
    """A robot that consists of a number of modules.

    It can be either a hardware robot or a simulation.
    """

    def __init__(self, modules: list[Module]) -> None:
        self.log = logging.getLogger(__name__)
        self.modules = modules


class RobotHardware(Robot):
    """A robot that consists of hardware modules.

    It generates Lizard code, forwards output to the hardware modules and sends commands to the robot brain.
    """

    def __init__(self, modules: list[Module], robot_brain: RobotBrain) -> None:
        super().__init__(modules)
        self.robot_brain = robot_brain
        self.robot_brain.lizard_code = self.generate_lizard_code()
        self.expander_prefixes = set(f'{module.name}:' for module in modules if isinstance(module, ExpanderHardware))
        rosys.on_repeat(self.update, 0.01)

    def generate_lizard_code(self) -> str:
        code = remove_indentation('''
            rdyp = Output(15)
            en3 = Output(12)
        ''')
        for module in self.modules:
            code += cast(ModuleHardware, module).lizard_code + '\n'
        output_fields = []
        for module in self.modules:
            output_fields.extend(cast(ModuleHardware, module).core_message_fields)
        code += remove_indentation(f'''
            core.output("core.millis {' '.join(output_fields)}")
            rdyp.on()
            en3.on()
        ''')
        return code

    async def update(self) -> None:
        for time, line in await self.robot_brain.read_lines():
            words = line.split()
            if not words:
                continue
            if words[0] in self.expander_prefixes:
                words.pop(0)
            if not words:
                continue
            if words[0] == 'core':
                words.pop(0)
                words.pop(0)
                for module in self.modules:
                    cast(ModuleHardware, module).handle_core_output(time, words)
            else:
                for module in self.modules:
                    if words[0] in cast(ModuleHardware, module).message_hooks:
                        cast(ModuleHardware, module).message_hooks[words[0]](line)


class RobotSimulation(Robot):
    """A robot that consists of simulated modules.

    It regularly calls the step method of all modules to allow them to update their internal state.
    """

    def __init__(self, modules: list[Module]) -> None:
        super().__init__(modules)
        self._last_step: Optional[float] = None
        rosys.on_repeat(self.step, 0.01)

    async def step(self) -> None:
        if self._last_step is not None:
            dt = rosys.time() - self._last_step
            for module in self.modules:
                await cast(ModuleSimulation, module).step(dt)
        self._last_step = rosys.time()



from pathlib import Path

import watchfiles
from nicegui import background_tasks, run, ui

VIDEO_FILES = Path('~/.rosys/timelapse/videos').expanduser()


class VideosPage:

    def __init__(self) -> None:
        @ui.page('/videos', title='Videos')
        async def page():
            @ui.refreshable
            async def videos():
                for mp4 in sorted(await run.io_bound(VIDEO_FILES.glob, '*.mp4'), reverse=True):
                    with ui.row():
                        with ui.card().tight():
                            video = ui.video(mp4).classes('w-[800px]')
                            src = video._props['src']  # pylint: disable=protected-access
                        with ui.column():
                            ui.button(on_click=lambda mp4=mp4: mp4.unlink()).props('icon=delete flat')
                            ui.button(on_click=lambda src=src: ui.download(src)).props('icon=download flat')

            async def watch_videos():
                async for _ in watchfiles.awatch(VIDEO_FILES):
                    videos.refresh()

            await videos()
            background_tasks.create(watch_videos(), name='watch videos files')



import abc
from typing import Optional

from ..event import Event
from .estop import EStop
from .expander import ExpanderHardware
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class Bumper(Module, abc.ABC):
    """A module that detects when a bumper is triggered."""

    def __init__(self, estop: Optional[EStop], **kwargs) -> None:
        self.estop = estop
        super().__init__(**kwargs)

        self.BUMPER_TRIGGERED = Event()
        """a bumper was triggered (argument: the bumper name)"""

        self.active_bumpers: list[str] = []


class BumperHardware(Bumper, ModuleHardware):
    """Hardware implementation of the bumper module.

    The module expects a dictionary of pin names and pin numbers.
    If an e-stop is provided, the module will not trigger bumpers if the e-stop is active.
    """

    def __init__(self, robot_brain: RobotBrain, *,
                 expander: Optional[ExpanderHardware] = None,
                 name: str = 'bumper',
                 pins: dict[str, int],
                 estop: Optional[EStop] = None) -> None:
        self.name = name
        self.pins = pins
        lizard_code = '\n'.join(f'{name}_{pin} = {expander.name + "." if expander else ""}Input({number})'
                                for pin, number in pins.items())
        core_message_fields = [f'{name}_{pin}.level' for pin in pins]
        super().__init__(robot_brain=robot_brain,
                         lizard_code=lizard_code,
                         core_message_fields=core_message_fields,
                         estop=estop)

    def handle_core_output(self, time: float, words: list[str]) -> None:
        if self.estop and self.estop.active:
            return
        active_bumpers = [pin for pin in self.pins if int(words.pop(0)) == 1]
        for pin in active_bumpers:
            if pin not in self.active_bumpers:
                self.BUMPER_TRIGGERED.emit(pin)
        self.active_bumpers[:] = active_bumpers


class BumperSimulation(Bumper, ModuleSimulation):
    """Simulation of the bumper module."""

    def set_active(self, pin: str, active: bool) -> None:
        if active and pin not in self.active_bumpers:
            self.BUMPER_TRIGGERED.emit(pin)
            self.active_bumpers.append(pin)
        if not active and pin in self.active_bumpers:
            self.active_bumpers.remove(pin)



import json
import time
from dataclasses import dataclass
from typing import Any, Optional

import requests


@dataclass
class JovisionCameraSettings:
    """Dataclass that holds the settings of an RTSP camera"""
    codec: str
    fps: int
    quality: str
    resolution: tuple[int, int]
    bitrate_control: str
    bitrate: int
    stream_id: int
    channel_id: int = 0

    def to_dict(self) -> dict[str, Any]:
        return {
            'channelid': self.channel_id,
            'streamid': self.stream_id,
            'venctype': self.codec,
            'width': self.resolution[0],
            'height': self.resolution[1],
            'frameRate': self.fps,
            'bitRate': self.bitrate,
            'ngop_s': 100,
            'quality': self.quality,
            'rcMode': self.bitrate_control,
        }


class JovisionInterface:
    """Interface that communicates with the RTSP camera settings mimicking the webpage requests"""

    def __init__(self, ip: str) -> None:
        self.ip = ip

    @property
    def settings_url(self) -> str:
        return f'http://{self.ip}/cgi-bin/jvsweb.cgi'

    @property
    def headers(self) -> dict[str, str]:
        return {'Cookie': 'passwdRule=1; username=admin; password=admin;'}

    def _send_settings(self, settings: list[JovisionCameraSettings]) -> None:
        """
        Set the settings for all streams
        """
        streams: list[dict] = []
        for i, setting in enumerate(settings):
            parameter_dict = setting.to_dict()
            parameter_dict['smartencode'] = 'open' if i < len(settings) - 1 else 'close'
            streams.append(parameter_dict)

        cmd = {
            'method': 'stream_set_params',
            'user': {
                'name': 'admin',
                'digest': '479ead4555b49227dc8812d06970c652',
            },
            'param': {
                'channelid': 0,
                'streams': streams,
            },
        }
        params = {
            'cmd': json.dumps(cmd),
            '_': int(time.time() * 1000),  # current time as a timestamp
        }
        requests.get(self.settings_url, params=params, headers=self.headers, timeout=1)  # type: ignore

    def set_fps(self, stream_id, fps) -> None:
        current_settings = self.get_current_settings()
        for settings in current_settings:
            if settings.stream_id == stream_id:
                settings.fps = fps
                break
        self._send_settings(current_settings)

    def get_fps(self, stream_id) -> Optional[int]:
        current_settings = self.get_current_settings()
        for settings in current_settings:
            if settings.stream_id == stream_id:
                return settings.fps

        return None

    def get_current_settings(self) -> list[JovisionCameraSettings]:
        cmd = {
            'method': 'stream_get_params',
            'user': {
                'name': 'admin',
                'digest': 'd065077d3e61cb56d45bd2dd28b83842',
            },
            'param': {
                'channelid': 0,
            },
        }
        params = {
            'cmd': json.dumps(cmd),
            '_': int(time.time() * 1000),
        }
        response = requests.get(self.settings_url, params=params, headers=self.headers, timeout=1)  # type: ignore

        return [
            JovisionCameraSettings(
                codec=stream['venctype'],
                fps=stream['frameRate'],
                quality=stream['quality'],
                resolution=(stream['width'], stream['height']),
                bitrate_control=stream['rcMode'],
                bitrate=stream['bitRate'],
                stream_id=stream['streamid'],
                channel_id=stream['channelid'],
            )
            for stream in response.json()['result']['streams']
        ]

    def get_parameter_ranges(self):
        raise NotImplementedError
        # pylint: disable=unreachable
        cmd = {
            'method': 'stream_get_all_ability',
            'user': {
                'name': 'admin',
                'digest': '8d5985ea2b9994aacb6cb3e3f826aae5'
            },
            'param': {
                'channelid': 0,
            }
        }
        params = {
            'cmd': json.dumps(cmd),
            '_': time.time() * 1000,
        }
        response = requests.get(self.settings_url, params=params, headers=self.headers, timeout=1)

        for stream_id, stream in enumerate(response.json()['result']['all']):
            print(f'stream {stream_id}')
            resolutions = stream['resolutions']
            for resolution in resolutions:
                width = resolution['width']
                height = resolution['height']
                b_default = resolution['bDefault']
                max_kbps = resolution['maxKbps']
                min_kbps = resolution['minKbps']
                def_kbps = resolution['defKbps']
                max_fr = resolution['maxFr']
                min_fr = resolution['minFr']
                def_fr = resolution['defFr']
            max_quality = stream['maxQuality']
            min_quality = stream['minQuality']
            max_ngop = stream['maxNGOP']
            min_ngop = stream['minNGOP']
            b_support_h265 = stream['bSupportH265']
            b_support_smart_enc = stream['bSupportSmartEnc']



import time

import yappi
from nicegui import ui
from tabulate import tabulate

from .. import rosys
from .profiling import profile


class ProfileButton(ui.button):
    """The profile button allows starting and stopping a profiling session.

    Use the `profiling.profile` decorator for including functions or methods in the analysis.
    The results are shown on the console.
    """

    def __init__(self) -> None:
        super().__init__('Profiler', on_click=self.toggle)
        self.props('icon=play_arrow')
        self.tooltip('run profiling')

    async def start(self, duration: float = 10.0) -> None:
        ui.notify('start profiling')
        self.props('icon=stop')
        yappi.clear_stats()
        yappi.start()
        profile.start()
        t = time.time()
        while yappi.is_running() and time.time() < t + duration:
            await rosys.sleep(0.1)
        self.stop()

    def stop(self) -> None:
        ui.notify('stop profiling')
        self.props('icon=play_arrow')
        profile.stop(print_stats=False)
        yappi.stop()
        table = [
            [str(v) for v in [stat.full_name, stat.ttot, stat.tsub, stat.tavg, stat.ncall]]
            for stat in yappi.get_func_stats()
            if 'python' not in stat.module
        ]
        output = tabulate(table[:15], headers=['function', 'total', 'excl. sub', 'avg', 'ncall'], floatfmt='.4f')
        print(output, flush=True)
        yappi.get_thread_stats().print_all()  # pylint: disable=no-member
        profile.print()

    async def toggle(self) -> None:
        if yappi.is_running():
            self.stop()
        else:
            await self.start()



import logging
from enum import Enum

import rosys
from rosys.event import Event

from .drivable import Drivable


class State(Enum):
    IDLE = 1
    STEERING = 2
    STOPPING = 3


class Steerer:
    """The steerer module translates x-y information (e.g. from a joystick) to linear/angular velocities sent to the robot.

    The wheels module can be any drivable hardware representation.
    Changing the steering state emits events that can be used to react to manual user interaction.
    """

    def __init__(self, wheels: Drivable, speed_scaling: float = 1.0) -> None:
        self.STEERING_STARTED = Event()
        """steering has started"""

        self.STEERING_STOPPED = Event()
        """steering has stopped"""

        self.log = logging.getLogger('rosys.steerer')

        self.wheels = wheels
        self.speed_scaling = speed_scaling
        self.state = State.IDLE
        self.linear_speed = 0.0
        self.angular_speed = 0.0

        rosys.on_repeat(self.step, 0.05)

    def start(self) -> None:
        self.log.info('start steering')
        self.state = State.STEERING
        self.STEERING_STARTED.emit()

    def update(self, x: float, y: float) -> None:
        if self.state == State.STEERING:
            self.linear_speed = y * self.speed_scaling
            self.angular_speed = -x * self.speed_scaling

    def stop(self) -> None:
        if self.state != State.STEERING:
            return
        self.log.info('stop steering')
        self.state = State.STOPPING
        self.STEERING_STOPPED.emit()

    async def step(self) -> None:
        if self.state == State.STEERING:
            await self.wheels.drive(self.linear_speed, self.angular_speed)
        elif self.state == State.STOPPING:
            await self.wheels.stop()
            self.state = State.IDLE

    def set_speed_scaling(self, speed_scaling: float) -> None:
        self.speed_scaling = speed_scaling

    def __str__(self) -> str:
        return f'{type(self).__name__} ({self.state})'



from dataclasses import dataclass
from typing import Optional


@dataclass(slots=True, kw_only=True)
class BmsState:
    percentage: Optional[float] = None
    voltage: Optional[float] = None
    current: Optional[float] = None
    temperature: Optional[float] = None
    is_charging: Optional[bool] = None
    last_update: float = 0

    @property
    def short_string(self) -> str:
        parts = []
        parts += [f'{self.percentage:.1f}%'] if self.percentage is not None else []
        parts += [f'{self.voltage:.1f}V'] if self.voltage is not None else []
        parts += [f'{self.current:.1f}A'] if self.current is not None else []
        parts += [f'{self.temperature:.1f}°C'] if self.temperature is not None else []
        return ', '.join(parts) or '?'



from typing import Optional

from .mjpeg_device import MjpegDevice
from .motec_settings_interface import MotecSettingsInterface
from .vendors import VendorType, mac_to_vendor


class MotecMjpegDevice(MjpegDevice):
    def __init__(self, mac: str, ip: str, *,
                 username: Optional[str] = '',
                 password: Optional[str] = '',
                 control_port: Optional[int] = 8885) -> None:

        super().__init__(mac, ip,
                         username=username, password=password)

        vendor = mac_to_vendor(mac)
        if vendor != VendorType.MOTEC:
            raise ValueError(f'MotecMjpegDevice can only be used with MOTEC devices. Got {vendor} for mac="{mac}"')

        self.settings_interface = MotecSettingsInterface(ip, port=control_port or 8885)

    async def set_fps(self, fps: int) -> None:
        await self.settings_interface.set_fps(fps)

    async def get_fps(self) -> int:
        return await self.settings_interface.get_fps()

    async def set_resolution(self, width: int, height: int) -> None:
        await self.settings_interface.set_stream_resolution(width, height)

    async def get_resolution(self) -> tuple[int, int]:
        return await self.settings_interface.get_stream_resolution()

    async def set_stream_port(self, port: int) -> None:
        await self.settings_interface.set_stream_port(port)

    async def get_stream_port(self) -> int:
        return await self.settings_interface.get_stream_port()

    async def get_stream_compression(self) -> int:
        return await self.settings_interface.get_stream_compression()

    async def set_stream_compression(self, level: int) -> None:
        await self.settings_interface.set_stream_compression(level)



import logging
from typing import Any, Optional

from typing_extensions import Self

from ... import rosys
from ..camera.configurable_camera import ConfigurableCamera
from ..camera.transformable_camera import TransformableCamera
from ..image import Image
from ..image_processing import get_image_size_from_bytes, process_jpeg_image
from .rtsp_device import RtspDevice


class RtspCamera(ConfigurableCamera, TransformableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: Optional[str] = None,
                 connect_after_init: bool = True,
                 streaming: bool = True,
                 fps: int = 5,
                 jovision_profile: int = 1,
                 ip: Optional[str] = None,
                 **kwargs,
                 ) -> None:
        super().__init__(id=id,
                         name=name,
                         connect_after_init=connect_after_init,
                         polling_interval=1.0 / fps,
                         streaming=streaming,
                         **kwargs)

        self.log = logging.getLogger(f'rosys.vision.rtsp_camera.{self.id}')

        self.device: Optional[RtspDevice] = None
        self.jovision_profile: int = jovision_profile
        self.ip: Optional[str] = ip

        self._register_parameter('fps', self.get_fps, self.set_fps,
                                 min_value=1, max_value=30, step=1, default_value=fps)
        self._register_parameter('jovision_profile', self.get_jovision_profile, self.set_jovision_profile,
                                 min_value=1, max_value=2, step=1, default_value=jovision_profile)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {
            name: param.value for name, param in self._parameters.items()
        } | {
            'ip': self.ip,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)

    @property
    def is_connected(self) -> bool:
        return self.device is not None and self.device.capture_task is not None

    @property
    def url(self) -> Optional[str]:
        if not self.is_connected:
            return None
        assert self.device is not None

        return self.device.url

    async def connect(self) -> None:
        if self.is_connected:
            return

        if not self.ip:
            self.log.error('no IP address provided for camera %s', self.id)
            return

        self.device = RtspDevice(mac=self.id, ip=self.ip, jovision_profile=self.jovision_profile)

        await self._apply_all_parameters()

    async def disconnect(self) -> None:
        if not self.is_connected:
            return
        logging.info('camera %s: disconnect initialized...', self.id)

        assert self.device is not None
        self.device.shutdown()
        self.device = None

    async def capture_image(self) -> None:
        if not self.is_connected:
            return
        assert self.device is not None

        image_bytes = self.device.capture()

        if not image_bytes:
            return
        transformed_image_bytes = await rosys.run.cpu_bound(process_jpeg_image, image_bytes, self.rotation, self.crop)
        if transformed_image_bytes is None:
            return

        try:
            final_image_resolution = get_image_size_from_bytes(transformed_image_bytes)
        except ValueError:
            return

        image = Image(time=rosys.time(), camera_id=self.id, size=final_image_resolution, data=transformed_image_bytes)
        self._add_image(image)

    def set_fps(self, fps: int) -> None:
        if self.device is None or self.device.settings_interface is None:
            return
        self.device.settings_interface.set_fps(stream_id=self.jovision_profile, fps=fps)
        self.polling_interval = 1.0 / fps

    def get_fps(self) -> Optional[int]:
        if self.device is None or self.device.settings_interface is None:
            return None
        fps = self.device.settings_interface.get_fps(stream_id=self.jovision_profile)
        return fps

    def set_jovision_profile(self, profile: int) -> None:
        if self.device is None:
            return
        self.jovision_profile = profile

    def get_jovision_profile(self) -> Optional[int]:
        if self.device is None:
            return None
        return self.jovision_profile

    async def _apply_parameters(self, new_values: dict[str, Any], force_set: bool = False) -> None:
        await super()._apply_parameters(new_values, force_set)
        if self.is_connected:
            assert self.device is not None
            self.device.restart_gstreamer()



from dataclasses import dataclass, field
from datetime import datetime

from rosys import persistence

from .image import Image


@dataclass(slots=True, kw_only=True)
class Uploads:
    minimal_minutes_between_uploads: float = 1.0
    last_upload: datetime = datetime.fromtimestamp(0)
    queue: dict[str, Image] = field(default_factory=dict, metadata=persistence.exclude)
    priority_queue: dict[str, Image] = field(default_factory=dict, metadata=persistence.exclude)

    def mark(self, image: Image, *, force: bool = False) -> None:
        """Mark an image for upload. Set force=True to circumvent minimal_minutes_between_uploads."""
        if force:
            self.priority_queue[image.id] = image
        else:
            self.queue[image.id] = image

    def get_queued(self) -> list[Image]:
        return [image for image in self.queue.values() if image.data]

    def get_priority_queued(self) -> list[Image]:
        return [image for image in self.priority_queue.values() if image.data]



import logging
import logging.config

import coloredlogs
import icecream

import rosys
from rosys.helpers import PackagePathFilter


class RosysFilter(logging.Filter):

    def filter(self, record: logging.LogRecord) -> bool:
        from rosys.testing.helpers import odometer as odo  # pylint: disable=import-outside-toplevel
        if odo:
            record.rosys_time = rosys.time()
            record.robot_pose = f'{odo.prediction.x:.2f}, {odo.prediction.y:1.2f}, {odo.prediction.yaw_deg:1.2f}'
        else:
            record.rosys_time = 0
            record.robot_pose = 'no robot pose yet'
        return True


def setup() -> None:
    icecream.install()

    config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'default': {
                '()': coloredlogs.ColoredFormatter,
                'format': r'%(rosys_time).2f [%(levelname)s] %(robot_pose)s %(relative_path)s:%(lineno)d: %(message)s',
                'datefmt': r'%Y-%m-%d %H:%M:%S',
            },
        },
        'filters': {
            'package_path_filter': {
                '()': PackagePathFilter,
            },
            'rosys_filter': {
                '()': RosysFilter,
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'filters': ['package_path_filter', 'rosys_filter'],
                'level': 'INFO',
                'stream': 'ext://sys.stdout'
            },
        },
        'loggers': {
            '': {  # root logger
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
            'rosys': {
                'handlers': ['console'],
                'level': 'DEBUG',
                'propagate': False,
            },
            'rosys.ota': {
                'handlers': ['console'],
                'level': 'WARN',
                'propagate': False,
            }
        },
    }

    logging.config.dictConfig(config)



import logging
from copy import deepcopy
from typing import Optional, Protocol

from .. import rosys
from ..event import Event
from ..geometry import Pose, PoseStep, Velocity


class VelocityProvider(Protocol):
    VELOCITY_MEASURED: Event


class Odometer:
    """An odometer collects velocity information from a given wheels module (or any velocity-providing hardware representation).

    It can also handle "detections", i.e. absolute pose information with timestamps.
    Given the history of previously received velocities, it can update its prediction of the current pose.

    The `get_pose` method provides robot poses from the within the last 10 seconds.
    """

    def __init__(self, wheels: VelocityProvider) -> None:
        self.ROBOT_MOVED = Event()
        """a robot movement is detected"""

        self.log = logging.getLogger('rosys.odometer')

        wheels.VELOCITY_MEASURED.register(self.handle_velocities)
        self.prediction: Pose = Pose()
        self.detection: Optional[Pose] = None
        self.current_velocity: Optional[Velocity] = None
        self.last_movement: float = 0
        self.history: list[Pose] = []
        self.odometry_frame: Pose = Pose()

        rosys.on_repeat(self.prune_history, 1.0)

    def handle_velocities(self, velocities: list[Velocity]) -> None:
        robot_moved: bool = False
        for velocity in velocities:
            if not self.history:
                self.history.append(Pose(time=velocity.time))
                continue

            dt = velocity.time - self.history[-1].time
            step = PoseStep(linear=dt*velocity.linear, angular=dt*velocity.angular, time=velocity.time)
            self.history.append(self.history[-1] + step)

            if step.linear or step.angular:
                robot_moved = True

        if self.history:
            self.prediction = self.odometry_frame.transform_pose(self.history[-1])
        if velocities:
            self.current_velocity = velocities[-1]
        if robot_moved:
            self.last_movement = step.time
            self.ROBOT_MOVED.emit()

    def handle_detection(self, detection: Pose) -> None:
        self.detection = detection

        if self.history:
            self.odometry_frame = self._compute_odometry_frame(self.get_pose(detection.time, local=True), detection)
            self.prediction = self.odometry_frame.transform_pose(self.history[-1])
        else:
            self.prediction = deepcopy(detection)

    def prune_history(self, max_age: float = 10.0) -> None:
        cut_off_time = rosys.time() - max_age
        while self.history and self.history[0].time <= cut_off_time:
            del self.history[0]

    def get_pose(self, time: float, local: bool = False) -> Pose:
        for i in range(1, len(self.history)):
            if self.history[i-1].time <= time < self.history[i].time:
                f = (time - self.history[i-1].time) / (self.history[i].time - self.history[i-1].time)
                local_pose = self.history[i-1].interpolate(self.history[i], f)
                return local_pose if local else self.odometry_frame.transform_pose(local_pose)
        if local:
            return deepcopy(self.history[-1])
        return Pose(x=self.prediction.x, y=self.prediction.y, yaw=self.prediction.yaw, time=time)

    @staticmethod
    def _compute_odometry_frame(local_pose: Pose, global_pose: Pose) -> Pose:
        frame = Pose.from_matrix(global_pose.matrix @ local_pose.inv_matrix)
        frame.time = global_pose.time
        return frame

    def reset(self) -> None:
        self.prediction = Pose()
        self.detection = None
        self.current_velocity = None
        self.last_movement = 0
        self.history.clear()
        self.odometry_frame = Pose()



import itertools
import logging
from dataclasses import dataclass
from functools import lru_cache
from typing import Optional

import networkx as nx
import numpy as np
from scipy import ndimage, spatial

from ..driving import PathSegment
from ..geometry import Point, Pose, PoseStep, Spline
from ..helpers import angle
from .area import Area
from .delaunay_pose_group import DelaunayPoseGroup
from .fast_spline import FastSpline
from .grid import Grid
from .obstacle import Obstacle
from .obstacle_map import ObstacleMap

GRID_RESOLUTION = 1.0
MIN_MARGIN = 1.0

TRY_SINGLE_PATH = True
"""Try to find a collision-free simple path between start and goal.

See https://trello.com/c/FtP4yHqA/777#comment-62d0323e97ba19392bcbceb8 for more information.
"""

TRY_SINGLE_SHUNTING = True
"""Try to find collision-free path with minimal switching (single shunting)."""


class DelaunayPlanner:

    def __init__(self, robot_outline: list[tuple[float, float]]) -> None:
        self.robot_outline = robot_outline
        self.areas: list[Area] = []
        self.obstacles: list[Obstacle] = []
        self.obstacle_map: Optional[ObstacleMap] = None
        self.tri_points: Optional[np.ndarray] = None
        self.tri_mesh: Optional[spatial.Delaunay] = None
        self.pose_groups: Optional[list[DelaunayPoseGroup]] = None
        self.graph: Optional[nx.DiGraph] = None
        self.log = logging.getLogger('rosys.delaunay_planner')

    def update_map(self, areas: list[Area], obstacles: list[Obstacle], additional_points: list[Point],
                   deadline: float) -> None:
        if self.obstacle_map and \
                self.areas == areas and \
                self.obstacles == obstacles and \
                all(self.obstacle_map.grid.contains(point, padding=1.0) for point in additional_points):
            return
        self.areas = areas
        self.obstacles = obstacles
        self._create_obstacle_map(additional_points, deadline)
        self._create_graph()

    def grow_map(self, points: list[Point], deadline: float) -> None:
        if self.obstacle_map is not None and \
                all(self.obstacle_map.grid.contains(point, padding=1.0) for point in points):
            return
        if self.obstacle_map is not None:
            bbox = self.obstacle_map.grid.bbox
            points.append(Point(x=bbox[0],         y=bbox[1]))
            points.append(Point(x=bbox[0]+bbox[2], y=bbox[1]))
            points.append(Point(x=bbox[0],         y=bbox[1]+bbox[3]))
            points.append(Point(x=bbox[0]+bbox[2], y=bbox[1]+bbox[3]))
        self._create_obstacle_map(points, deadline)
        self._create_graph()

    def _create_obstacle_map(self, additional_points: list[Point], deadline: float) -> None:
        points = [p for obstacle in self.obstacles for p in obstacle.outline]
        points += [p for area in self.areas for p in area.outline]
        points += additional_points
        grid = Grid.from_points(points, pixel_size=0.1, num_layers=36, padding=1.0)
        self.obstacle_map = ObstacleMap.from_world(self.robot_outline, self.areas, self.obstacles, grid, deadline)

    def _create_graph(self) -> None:
        assert self.obstacle_map is not None
        min_x, min_y, size_x, size_y = self.obstacle_map.grid.bbox
        X, Y = np.meshgrid(np.arange(min_x, min_x + size_x - GRID_RESOLUTION / 2, GRID_RESOLUTION),
                           np.arange(min_y, min_y + size_y, GRID_RESOLUTION * np.sqrt(3) / 2))
        X[::2] += GRID_RESOLUTION / 2

        rows, cols = self.obstacle_map.grid.to_grid(X.flatten(), Y.flatten())
        distance = ndimage.distance_transform_edt(1 - self.obstacle_map.map) * self.obstacle_map.grid.pixel_size
        D = ndimage.map_coordinates(distance, [[rows], [cols]], order=0).reshape(X.shape)
        gradient_y, gradient_x = np.gradient(distance)
        dD_dX = ndimage.map_coordinates(gradient_x, [[rows], [cols]], order=0).reshape(X.shape)
        dD_dY = ndimage.map_coordinates(gradient_y, [[rows], [cols]], order=0).reshape(X.shape)
        dD = np.sqrt(dD_dX**2 + dD_dY**2)
        close = np.logical_and(0.0 < D, D < MIN_MARGIN)
        close = np.logical_and(close, dD > 0)
        X[close] += dD_dX[close] / dD[close] * (MIN_MARGIN - D[close])
        Y[close] += dD_dY[close] / dD[close] * (MIN_MARGIN - D[close])

        keep = np.reshape([not all(self.obstacle_map.stack[int(np.round(row)), int(np.round(col)), :])
                           for row, col in zip(rows, cols)], X.shape)
        keep[1::2, :] = np.logical_and(keep[1::2, :], D[1::2, :] < 2)
        keep[::4, 1::2] = np.logical_and(keep[::4, 1::2], D[::4, 1::2] < 2)
        keep[2::4, ::2] = np.logical_and(keep[2::4, ::2], D[2::4, ::2] < 2)
        self.tri_points = np.stack((X[keep], Y[keep]), axis=1)
        assert self.tri_points is not None  # NOTE: mypy doesn't seem to understand np.stack

        self.tri_mesh = spatial.Delaunay(self.tri_points)
        self.pose_groups = [
            DelaunayPoseGroup(
                index=i,
                point=Point(x=self.tri_points[i, 0], y=self.tri_points[i, 1]),
                neighbor_indices=_tri_neighbors(self.tri_mesh, i).tolist(),
                poses=[
                    Pose(x=point[0], y=point[1], yaw=np.arctan2(neighbor[1] - point[1], neighbor[0] - point[0]))
                    for neighbor in self.tri_points[_tri_neighbors(self.tri_mesh, i)]
                ],
            )
            for i, point in enumerate(self.tri_points)
        ]

        self.graph = nx.DiGraph()
        for g, group in enumerate(self.pose_groups):
            for p in range(len(group.poses)):
                self.graph.add_node((g, p))
        for g, group in enumerate(self.pose_groups):
            for p, (pose, g_) in enumerate(zip(group.poses, group.neighbor_indices)):
                for p_, pose_ in enumerate(self.pose_groups[g_].poses):
                    if abs(angle(pose.yaw, pose_.yaw + np.pi)) < 0.01:
                        continue  # NOTE: avoid 180-degree turns
                    x, y, yaw = _generate_poses(self.obstacle_map.grid, pose, pose_)
                    if not self.obstacle_map.test(x, y, yaw).any():
                        length = np.sum(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
                        self.graph.add_edge((g, p), (g_, p_), backward=False, weight=length)
                        if ((g_, p_), (g, p)) not in self.graph.edges:
                            self.graph.add_edge((g_, p_), (g, p), backward=True, weight=1.2*length)

    def search(self, start: Pose, goal: Pose) -> list[PathSegment]:
        assert self.obstacle_map is not None
        assert self.graph is not None
        assert self.pose_groups is not None
        paths: list[list[PathSegment]] = []

        if TRY_SINGLE_PATH:
            for backward in [True, False]:
                simple_spline = Spline.from_poses(start, goal, backward=backward)
                if _is_healthy(simple_spline) and not self.obstacle_map.test_spline(simple_spline, backward):
                    paths.append([PathSegment(spline=simple_spline, backward=backward)])
        if paths:
            self.log.info('found single spline to reach goal')
            return min(paths, key=lambda path: path[0].spline.estimated_length())

        if TRY_SINGLE_SHUNTING:
            for backward in [True, False]:
                for length in [1, 1.5, 2]:
                    y_outline = [p[1] for p in self.robot_outline]
                    robot_length = max(y_outline) - min(y_outline)
                    step = PoseStep(linear=robot_length * (-length if backward else length), angular=0, time=0)
                    intermediate = start + step
                    spline1 = Spline.from_poses(start, intermediate, backward=backward)
                    if not _is_healthy(spline1) or self.obstacle_map.test_spline(spline1, backward):
                        continue
                    spline2 = Spline.from_poses(intermediate, goal, backward=not backward)
                    if not _is_healthy(spline2) or self.obstacle_map.test_spline(spline2, not backward):
                        continue
                    paths.append([
                        PathSegment(spline=spline1, backward=backward),
                        PathSegment(spline=spline2, backward=not backward),
                    ])
        if paths:
            self.log.info('found single shunt to reach goal')
            return min(paths, key=lambda path: path[0].spline.estimated_length() + path[1].spline.estimated_length())

        grid_entries = _find_grid_passages(self.obstacle_map, self.pose_groups, start, True)
        grid_exits = _find_grid_passages(self.obstacle_map, self.pose_groups, goal, False)
        if not grid_entries:
            raise RuntimeError('could not find start segment')
        if not grid_exits:
            raise RuntimeError('could not find exit segment')

        for enter, exit_ in itertools.product(grid_entries, grid_exits):
            p, g = enter.coordinate
            p_, g_ = exit_.coordinate
            path: list[PathSegment] = [enter.segment]
            last_g, last_p = g, p
            try:
                for next_g, next_p in nx.shortest_path(self.graph, (g, p), (g_, p_), weight='weight')[1:]:
                    last_pose = self.pose_groups[last_g].poses[last_p]
                    next_pose = self.pose_groups[next_g].poses[next_p]
                    backward = self.graph.edges[((last_g, last_p), (next_g, next_p))]['backward']
                    spline = Spline.from_poses(last_pose, next_pose, backward=backward)
                    path.append(PathSegment(spline=spline, backward=backward))
                    last_g, last_p = next_g, next_p
            except nx.exception.NetworkXNoPath:
                continue
            path.append(exit_.segment)

            while True:
                shortcuts: list[PathSegment] = []
                for step_size in [1, 2]:
                    for s in range(len(path) - step_size):
                        new_start = Pose(
                            x=path[s].spline.start.x,
                            y=path[s].spline.start.y,
                            yaw=path[s].spline.yaw(0) + (np.pi if path[s].backward else 0),
                        )
                        new_end = Pose(
                            x=path[s+step_size].spline.end.x,
                            y=path[s+step_size].spline.end.y,
                            yaw=path[s+step_size].spline.yaw(1) + (np.pi if path[s+step_size].backward else 0),
                        )
                        if abs(angle(new_start.yaw, new_end.yaw + np.pi)) < 0.01:
                            continue
                        for new_backward in [False, True]:
                            new_spline = Spline.from_poses(new_start, new_end, backward=new_backward)
                            if not _is_healthy(new_spline):
                                continue
                            if self.obstacle_map.test_spline(new_spline, new_backward):
                                continue
                            combined_length = path[s].spline.estimated_length() + \
                                path[s+step_size].spline.estimated_length()
                            if .9 * new_spline.estimated_length() > combined_length:
                                continue
                            shortcuts.append(PathSegment(spline=new_spline, backward=new_backward))
                        if shortcuts:
                            lengths = [segment.spline.estimated_length() for segment in shortcuts]
                            path[s] = shortcuts[np.argmin(lengths)]
                            for _ in range(step_size):
                                del path[s+1]
                            break  # restart while loop
                    if shortcuts:
                        break  # restart while loop
                if not shortcuts:
                    break  # exit while loop
            paths.append(path)
        if not paths:
            raise RuntimeError('could not find path')
        return min(paths, key=len)


def _tri_neighbors(tri_mesh: spatial.Delaunay, vertex_index: int) -> np.ndarray:
    return tri_mesh.vertex_neighbor_vertices[1][tri_mesh.vertex_neighbor_vertices[0][vertex_index]:
                                                tri_mesh.vertex_neighbor_vertices[0][vertex_index + 1]]


@lru_cache(maxsize=1000)
def _t_array(n: int) -> np.ndarray:
    return np.linspace(0, 1, n)


@lru_cache(maxsize=10000)
def _generate_pose_offsets(grid: Grid, dx: float, dy: float, yaw: float, yaw_: float) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    spline = FastSpline(0, 0, yaw, dx, dy, yaw_, False)
    row0, col0, layer0 = grid.to_3d_grid(0, 0, yaw)
    row1, col1, layer1 = grid.to_3d_grid(dx, dy, yaw_)
    num_rows = int(abs(row1 - row0))
    num_cols = int(abs(col1 - col0))
    num_layers = int(abs(layer1 - layer0))
    t = _t_array(max(num_rows, num_cols, num_layers))
    return (spline.x(t), spline.y(t), spline.yaw(t) + [0, np.pi][spline.backward])


def _generate_poses(grid: Grid, pose: Pose, pose_: Pose) -> tuple:
    dx, dy, yaw = _generate_pose_offsets(grid, pose_.x - pose.x, pose_.y - pose.y, pose.yaw, pose_.yaw)
    return pose.x + dx, pose.y + dy, yaw


def _is_healthy(spline: Spline, curvature_limit: float = 10.0) -> bool:
    return np.abs(spline.max_curvature()) < curvature_limit


@dataclass(slots=True, kw_only=True)
class Passage:
    segment: PathSegment
    coordinate: tuple[int, int]


def _find_grid_passages(obstacle_map: ObstacleMap,
                        pose_groups: list[DelaunayPoseGroup],
                        pose: Pose,
                        entering: bool,
                        max_num_groups: int = 10,
                        max_num_results: int = 3) -> list[Passage]:
    group_distances = [g.point.distance(pose) for g in pose_groups]
    group_indices = np.argsort(group_distances)
    results: list[Passage] = []
    for g, group in zip(group_indices, np.array(pose_groups)[group_indices][:max_num_groups]):
        for p, group_pose in enumerate(group.poses):
            for backward in [False, True]:
                poses = (pose, group_pose) if entering else (group_pose, pose)
                spline = Spline.from_poses(*poses, backward=backward)
                if _is_healthy(spline) and not obstacle_map.test_spline(spline, backward):
                    passage = Passage(segment=PathSegment(spline=spline, backward=backward), coordinate=(p, g))
                    results.append(passage)
    results.sort(key=lambda passage: passage.segment.spline.estimated_length())
    return results[:max_num_results]



#!/usr/bin/env python
import pylab as pl
from nicegui import ui

from rosys.geometry import Pose, Spline
from rosys.pathplanning import plot_tools as pt
from rosys.pathplanning.grid import Grid
from rosys.pathplanning.obstacle_map import ObstacleMap
from rosys.pathplanning.robot_renderer import RobotRenderer

with ui.pyplot():
    pl.axis('equal')

    grid = Grid((3, 4, 36), (0.5, 1.5, 4.0, 3.0))
    o = [2.55, 2.60, 0.9, 0.89]
    robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)
    obstacle_map = ObstacleMap.from_list(grid, [o], robot_renderer)
    pt.plot_bbox(o, 'C3')
    pt.show_obstacle_map(obstacle_map)
    pl.gca().invert_yaxis()

    spline = Spline.from_poses(Pose(x=3, y=4, yaw=0.5), Pose(x=1, y=2, yaw=0), control_dist=0.5, backward=True)
    pl.plot(spline.start.x, spline.start.y, 'C1o')
    pl.plot(spline.end.x, spline.end.y, 'C1o')
    pl.plot(spline.control1.x, spline.control1.y, 'C2s')
    pl.plot(spline.control2.x, spline.control2.y, 'C2s')
    pt.plot_spline(spline, backward=True)
    pt.plot_robot(robot_renderer, (spline.x(0), spline.y(0), spline.yaw(0)), lw=2)

    x, y = 2, 2
    pl.plot(x, y, 'C4o')
    t = spline.closest_point(x, y)
    pl.plot(spline.x(t), spline.y(t), 'C4x')

ui.run()



import asyncio
import logging

from nicegui import ui

from .. import rosys


class AsyncioWarnings:

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.asyncio_warnings')
        rosys.on_startup(lambda: self.set_threshold(0.05))

    def activate(self) -> None:
        """Produce warnings for coroutines which take too long on the main loop and hence clog the event loop.

        Beware: This slows everything down. Do not use in production or while profiling!
        """
        try:
            loop = asyncio.get_running_loop()
            loop.set_debug(True)
            self.log.warning('activated asyncio warnings; do not use in production or while profiling)')
        except Exception:
            self.log.exception('could not activate asyncio warnings')

    def deactivate(self) -> None:
        """Stop the warnings for slow coroutines."""
        try:
            loop = asyncio.get_running_loop()
            loop.set_debug(False)
        except Exception:
            self.log.exception('could not deactivate asyncio warnings')

    def toggle(self) -> None:
        loop = asyncio.get_running_loop()
        if loop.get_debug():
            self.deactivate()
        else:
            self.activate()

    def set_threshold(self, seconds: float) -> None:
        """Sets the threshold in seconds after which a coroutine is considered slow."""
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = seconds

    def ui(self) -> ui.element:
        with ui.row().classes('items-end') as content:
            loop = asyncio.get_running_loop()
            ui.checkbox(on_change=self.toggle)
            ui.number('coro limit',
                      value=loop.slow_callback_duration, format='%.3f',
                      on_change=lambda e: self.set_threshold(e.value)) \
                .props('suffix=s').classes('w-16')
        return content
